//==============================================================================
// This file is PUBLIC DOMAIN
//==============================================================================
// includes

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>


//==============================================================================
// CRC24

#include "CRC24.h"

//==============================================================================
// debugging defines

// #define DEBUG_6 1

#if DEBUG_6 != 0
FILE *dFile6;
#endif

//==============================================================================
// defines

#define STR_LENGTH 2048
#define LAST_GLOBAL 0x415
#define NUM_WANTS 628
#define NUM_OBJECTS 3314
// #define NUM_PGROUPS 1261

typedef struct tagPackRec {
    unsigned long *toc;
    unsigned long size;
    unsigned long ver;
    char name[STR_LENGTH];
} PackRec, *pPackRec;

//==============================================================================
// declarations

void loadToc(char *s);
void readGx(unsigned char c);
void readGx2(unsigned short w);
void data2(unsigned char o, unsigned short x);
void readGUID(unsigned long ID);
void readO2(unsigned short opcode, unsigned char *b, unsigned long x, unsigned short f, unsigned char nodeVersion);
void read2(unsigned char *b, unsigned long s);
unsigned char *unPack(unsigned char *b, unsigned long s1, unsigned long s2);
unsigned char *get2(unsigned long o, unsigned long *s);
void readConst2(unsigned short instance, unsigned short id);
int readString2(unsigned long group, unsigned long instance, unsigned short id);
int readFn2(unsigned short o);
void findFirst();
void findNext2(unsigned long g, unsigned long i, unsigned long *o1, unsigned long *s1);
void dis();

//==============================================================================
// global variables

PackRec gPack;
FILE *outFile;
unsigned long gGroup, gInstance, gNum;
unsigned char gNref[STR_LENGTH], gGlob[STR_LENGTH];
unsigned long gNrefCrc, gGlobGroup;

//==============================================================================
// data goes here

#include "disdata2.h"

//==============================================================================
// load toc and put data in my pack

void loadToc(char *s)
{
    FILE *f;
    int status = 0;
    unsigned long xStart, xSize, xVer, *pToc;
    pPackRec p3;

    // get the data for this file

    if ((f = fopen(s, "rb")) != NULL) {
        fseek(f, 0x28, 0);
        fread(&xStart, 4, 1, f);
        fread(&xSize, 4, 1, f);
        fseek(f, 0x3C, 0);
        fread(&xVer, 4, 1, f);              // subversion (1/2)

        if (xSize != 0) {
            fseek(f, xStart, 0);
            pToc = (unsigned long *) malloc(xSize);
            if (pToc != NULL) {
                fread(pToc, xSize, 1, f);   // load toc
                status = 1;
            }
        }
        fclose(f);
    }

    if (status) {
        gPack.toc = pToc;
        gPack.size = xSize;
        gPack.ver = xVer;
        strcpy(gPack.name, s);
    }
}

//==============================================================================
// read the specified goto code

void readGx(unsigned char c)
{
    switch (c) {
        case 0xFD:
            fprintf(outFile,"error");
            break;
        case 0xFE:
            fprintf(outFile,"true");
            break;
        case 0xFF:
            fprintf(outFile,"false");
            break;
        default:
            fprintf(outFile,"%X", c);
    }
}

void readGx2(unsigned short w)
{
    switch (w) {
        case 0xFFFC:
            fprintf(outFile,"error");
            break;
        case 0xFFFD:
            fprintf(outFile,"true");
            break;
        case 0xFFFE:
            fprintf(outFile,"false");
            break;
        default:
            fprintf(outFile,"%X", w);
    }
}

//==============================================================================
// read the decoded indicated data

void data2(unsigned char o, unsigned short x)
{
    unsigned short a, b, c;
    unsigned long d1, d2;

#if 0
    fprintf(stderr, "Group = %X, Instance = %X, %X %X     \r", gGroup, gInstance, o, x);
#endif

    switch(o) {     // data owner
        case 0:     // My Attribute
        case 1:     // Stack Obj's Attribute
            fprintf(outFile,"%s ", gDataOwners[o]);
            b = readString2(gGroup, 0x100, x);
            if (b) fprintf(outFile,"(");
            if (x < 10)
                fprintf(outFile,"%d", x);
            else
                fprintf(outFile,"%d (0x%X)", (short) x, x);
            if (b) fprintf(outFile,")");
            break;
        case 2:     // My Semi Attribute
        case 5:     // Stack Objects Semi Attribute
            fprintf(outFile,"%s ", gDataOwners[o]);
            b = readString2(gGlobGroup, 0x100, x);
            if (b) fprintf(outFile,"(");
            if (x < 10)
                fprintf(outFile,"%d", x);
            else
                fprintf(outFile,"%d (0x%X)", (short) x, x);
            if (b) fprintf(outFile,")");
            break;
        case 3:     // My
        case 4:     // Stack Object's
            if (x == 0xB) {     // object's object ID (itself)
                if (o == 3) fprintf(outFile,"Me");
                if (o == 4) fprintf(outFile,"Stack Object");
            } else
                fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gDataLabels[x], x);
            break;
        case 6:     // Global (from Simulation)
            fprintf(outFile,"Global(%s)", gGlobalLabels[x]);
            break;
        case 7:     // Literal Value
            if (x < 16)
                fprintf(outFile,"%d", x);
            else
                fprintf(outFile,"%d (0x%X)", (short) x, x);
            break;
        case 8:     // Temporary Storage
        case 9:     // Parameters
        case 0x19:  // Local Variable
            fprintf(outFile,"%s %d", gDataOwners[o], x);
            break;
        case 0x0A:  // Stack Object ID
            if (x == 0)
                fprintf(outFile,"%s", gDataOwners[o]);
            else
                fprintf(outFile,"invalid", x);
            break;
                    // data owner [temp x] (temp x = data index)
        case 0xB:   // Temporary Storage[temp]
        case 0x11:  // stack obj's motive[temp]
        case 0x1E:  // my person data [temp]
        case 0x1F:  // stack obj's person data [temp]
        case 0x30:  // My Slot [temp]
        case 0x31:  // Stack Objects Slot [temp]
            fprintf(outFile,"%s [Temp %d]", gDataOwners[o], x);
            break;
        case 0x0C:  // check tree ad range
        case 0x0E:  // my motives
        case 0x0F:  // stack obj's motives
        case 0x1C:  // check tree ad personality var
        case 0x1D:  // check tree ad min
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gMotives[x], x);
            break;
        case 0x12:  // my person data
        case 0x13:  // stack obj's person data
        case 0x20:  // neighbor's person data
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gPersonDataLabels[x], x);
            break;
        case 0x15:  // stack object's definition
        case 0x26:  // Neighbor's Object Definition
        case 0x33:  // Stack Object's Master Definition
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gObjectDefs[x], x);
            break;
        case 0x17:  // room [temp 0]
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gRoomValues[x], x);
            break;
        case 0x18:  // neighbor in stack object
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gNDataLabels[x], x);
            break;
        case 0x1A:  // Constant Value
            a = x >> 13;            // x = aa*bbbbb bccccccc
            b = (x >> 7) & 0x7F;
            c = x & 0x7F;
            switch (a) {
                case 0:             // private
                    b += 0x1000;
                    break;
                case 1:             // semi-global
                    b = 0x2000 + ((x >> 7) & 0x3F);
                    break;
                case 2:             // global
                    b += 0x100;
                    break;
            }
            fprintf(outFile,"%s 0x%X:%X", gDataOwners[o], b, c);
            readConst2(b, c);
            break;
        case 0x2F:  // Constants [temp]
            a = x >> 13;            // x = aa*bbbbb bbbbbccc
            b = (x >> 3) & 0x7FF;
            c = x & 7;
            switch (a) {
                case 0:             // private
                    b += 0x1000;
                    break;
                case 1:             // semi-global
                    b = 0x2000 + ((x >> 3) & 0x3FF);
                    break;
                case 2:             // global
                    b += 0x100;
                    break;
            }
            fprintf(outFile,"%s 0x%X:[Temp %d]", gDataOwners[o], b, c);
            break;
        case 0x21:  // job data [temp 0,1]
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gJobData[x], x);
            break;
        case 0x22:  // neighborhood data
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gNHDataLabels[x], x);
            break;
        case 0x23:  // stack object's function
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gEntryPoints[x], x);
            break;
        case 0x27:  // Temporary Token
        case 0x28:  // Stack Object's Temporary Token
            fprintf(outFile,"%s %s (0x%X)", gDataOwners[o], gInventoryDialogStrings[x], x);
            break;
        default:
            if (o > 0x33)
                fprintf(outFile,"*** Error: bad owner (0x%X, 0x%X) ***", o, x);
            else {
                if (x < 10)
                    fprintf(outFile,"%s %d", gDataOwners[o], x);
                else
                    fprintf(outFile,"%s %d (0x%X)", gDataOwners[o], (short) x, x);
            }
    }
}

//==============================================================================
// read the object name for the GUID if possible

void readGUID(unsigned long d1)
{
    unsigned short w1, w2;

    w2 = 0xFFFF;
    for (w1 = 0; w1 < NUM_OBJECTS; w1++) if (d1 == gObjectIDs[w1]) w2 = w1;
    if (w2 != 0xFFFF) fprintf(outFile," (%s)", gObjectNames[w2]);
}

//==============================================================================
// interpret the operands of the specified opcode
// ( where the REAL work is done... )

void readO2(unsigned short opcode, unsigned char *b, unsigned long x, unsigned short f, unsigned char nodeVersion)
{
    unsigned long d1;
    unsigned short w1, w2, w3, w4, w5;
    unsigned char c1, c2, c3;
    int i;

    if (opcode < 0x7E) {    // primitive
        if (opcode != 2 && opcode != 8 && opcode != 0xB && opcode != 0xC && opcode != 0x16 && opcode != 0x1E)
            fprintf(outFile,"%s: ", gPrims[opcode]);    // implied function opcodes
        if (*(unsigned long *) (&b[x]) != 0xFFFFFFFF) { // -1 = param killer
            switch (opcode) {
                case 0x00:  // Sleep (false = error)
                    fprintf(outFile,"for ");
                    data2(9, b[x]);
                    fprintf(outFile," ticks");
                    break;
                case 0x01:  // Generic Sims Call
                    c1 = b[x];
                    fprintf(outFile,"%s", gGenericCalls[c1]);
                    switch (c1) {
                        case 0:
                            fprintf(outFile," Temp 0:neighborhood, Temp 1:evict, Temp 2:save lot, Temp 3:reset tutorial");
                            break;
                        case 4:
                            fprintf(outFile," Stack Obj:nID, Temp 0:familyID");
                            break;
                        case 5:
                            fprintf(outFile," Temp 0:familyID");
                            break;
                        case 6:
                            fprintf(outFile," Stack Obj:nID");
                            break;
                        case 0x0D:
                            fprintf(outFile," Stack Obj");
                            break;
                        case 0x11:
                            fprintf(outFile," Temp 0:lotID");
                            break;
                        case 0x12:
                            fprintf(outFile," Temp 0:speed");
                            break;
                        case 0x15:
                            fprintf(outFile," Temp 0:child nID, Temp 1:parent nID");
                            break;
                        case 0x16:
                            fprintf(outFile," Temp 0:new spouse nID, Temp 1:initial spouse nID");
                            break;
                        case 0x17:
                            fprintf(outFile," Temp 0:remove nID, Temp 1:relative nID");
                            break;
                        case 0x18:
                            fprintf(outFile," Temp 0:age");
                            break;
                        case 0x19:
                        case 0x28:
                        case 0x29:
                        case 0x2C:
                        case 0x31:
                        case 0x33:
                            fprintf(outFile," Temp 0");
                            break;
                        case 0x1C:
                            fprintf(outFile," Temp 0:amount, Temp 2:multiplier");
                            break;
                        case 0x1D:
                        case 0x26:
                            fprintf(outFile," Temp 0:nID");
                            break;
                        case 0x1E:
                            fprintf(outFile," Temp 0:outfit");
                            break;
                        case 0x1F:
                            fprintf(outFile," Temp 0,1:GUID");
                            break;
                        case 0x20:
                            fprintf(outFile," Temp 0:outfit, Temp 1:result value");
                            break;
                        case 0x22:
                            fprintf(outFile," Temp 0:source, Temp 1:destination, Temp 2:result value");
                            break;
                        case 0x23:
                        case 0x24:
                            fprintf(outFile," Temp 0:lotID, Temp 1:result value");
                            break;
                        case 0x2D:
                            fprintf(outFile," Temp 0:source, Stack Obj:destination");
                            break;
                        case 0x2E:
                            fprintf(outFile," Temp 0:tableID, Temp 1:index, Temp 3:fallback");
                            break;
                        case 0x30:
                            fprintf(outFile," Temp 0:nID, Temp 1:result value");
                            break;
                        case 0x32:
                            fprintf(outFile," Temp 0:take, Temp 1:target nID, Temp 2:percent, Temp 3,4:amount, Temp 5:from assets");
                            break;
                        case 0x34:
                            fprintf(outFile," Temp 0:direction, Temp 1:result wall obj");
                            break;
                        case 0x35:
                            fprintf(outFile," Temp 0:tableID, Temp 1:index, Temp 2:state, Temp 3:fallback");
                            break;
                    }
                    break;
                case 0x02:  // (Evaluate) Expression (non-comparison operators return false if error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    c1 = b[x+5];
                    c2 = b[x+6];
                    data2(c2, w1);            // target data
                    if (c1 == 0x14)           // abs(rhs)
                        fprintf(outFile," := abs(");
                    else if (c1 == 0x15)      // Assign 32bit Value (both target and source are contiguous 32-bit)
                        fprintf(outFile," := int32(");
                    else
                        fprintf(outFile," %s ",gOperators[c1]);

                    if (b[x+7] == 7 && c1 > 7 && c1 < 11 && w2 > 0) {   // literal (non-0), flag operator
                        if ((c2 == 3 || c2 == 4) && (w1 == 4 || w1 == 5 || w1 == 8 ||
                                w1 == 0x0D || w1 == 0x22 || w1 == 0x28 || w1 == 0x2A ||
                                w1 == 0x2B || w1 == 0x3F || w1 == 0x45)) {
                            switch (w1) {
                                case 4:       // allowed height flags
                                    fprintf(outFile,"%s", gAllowedHeightFlags[w2 - 1]);
                                    break;
                                case 5:       // wall adj. flags
                                    fprintf(outFile,"%s", gWallAdjFlags[w2 - 1]);
                                    break;
                                case 8:       // flags
                                    fprintf(outFile,"%s", gFlags1[w2 - 1]);
                                    break;
                                case 0x0D:    // wall placement flags
                                    fprintf(outFile,"%s", gWallFlags[w2 - 1]);
                                    break;
                                case 0x22:    // Hidden Flags
                                    fprintf(outFile,"%s", gHiddenFlags[w2 - 1]);
                                    break;
                                case 0x28:    // flags 2
                                    fprintf(outFile,"%s", gFlags2[w2 - 1]);
                                    break;
                                case 0x2A:    // placement flags
                                    fprintf(outFile,"%s", gHeightFlags[w2 - 1]);
                                    break;
                                case 0x2B:    // movement flags
                                    fprintf(outFile,"%s", gMoveFlags[w2 - 1]);
                                    break;
                                case 0x3F:    // exclusive placement flags
                                    fprintf(outFile,"%s", gPlacementFlags[w2 - 1]);
                                    break;
                                case 0x45:    // wall cutout flags
                                    fprintf(outFile,"%s", gWallCutoutFlags[w2 - 1]);
                                    break;
                            }
                        } else if ((c2 == 0x12 || c2 == 0x13 || c2 == 0x20) &&
                        (w1 == 0x1E || w1 == 0x44 || w1 == 0x51 || w1 == 0x9E || w1 == 0x9F)) {
                            switch (w1) {
                                case 0x1E:    //censorship flags
                                    fprintf(outFile,"%s", gCensorFlags[w2 - 1]);
                                    break;
                                case 0x44:    //Ghost Flags
                                    fprintf(outFile,"%s", gGhostFlags[w2 - 1]);
                                    break;
                                case 0x51:    //Body Flags
                                    fprintf(outFile,"%s", gBodyFlags[w2 - 1]);
                                    break;
                                case 0x9E:    //Selection Flags
                                    fprintf(outFile,"%s", gSelectionFlags[w2 - 1]);
                                    break;
                                case 0x9F:    //Person Flags
                                    fprintf(outFile,"%s", gPersonFlags[w2 - 1]);
                                    break;
                            }
                        } else if ((c2 == 0x15 || c2 == 0x26 || c2 == 0x33) &&
                        (w1 == 0x27 || w1 == 0x28)) {
                            switch (w1) {
                                case 0x27:    //room sort flags
                                    fprintf(outFile,"%s", gRoomSortFlags[w2 - 1]);
                                    break;
                                case 0x28:    //function sort flags
                                    fprintf(outFile,"%s", gFunctionSortFlags[w2 - 1]);
                                    break;
                            }
                        } else {
                            data2(b[x+7], w2);  // unknown flag
                        }
                    } else {
                        data2(b[x+7], w2);      // not a flag
                        if (c1 >= 0x14)         // abs(rhs) or Assign 32bit Value
                            fprintf(outFile,")");
                    }
                    break;
                case 0x03:  // Find Best Interaction
                    w1 = *(unsigned short *) (&b[x]);   // motive
                    w2 = *(unsigned short *) (&b[x+2]); // flags
                    if (w2 == 0)
                        fprintf(outFile,"for the N current worst motives.");
                    else {
                        w3 = 1;
                        for (c1 = 0; c1 < 16; c1++) {   // this should only find 1 motive (if any)
                            if (w1 & w3) {
                                fprintf(outFile," for %s", gMotives[c1]);;
                            }
                            w3 = w3 << 1;
                        }
                        fprintf(outFile,".");
                        if (w2 & 1)
                            fprintf(outFile," Choose remaining motives from the worst.");
                    }
                    if (w2 & 0x200) {
                        fprintf(outFile,", Only test objects in room held in temp 0");
                        if (w2 & 0x400)
                            fprintf(outFile," including Out Of World Objects");
                        fprintf(outFile,".");
                    }
                    if (w2 & 0x800) {
                        fprintf(outFile,", Find Nested Interactions Only");
                        if (w2 & 0x1000)
                            fprintf(outFile," and only on current Interaction object");
                        fprintf(outFile,".");
                    }
                    break;
                case 0x07:  // Refresh (false = error)
                    if (b[x] == 0)
                        fprintf(outFile,"My ");
                    else
                        fprintf(outFile,"Stack Object's ");
                    if (b[x+2] == 0)
                        fprintf(outFile,"graphic");
                    else if (b[x+2] == 1)
                        fprintf(outFile,"lighting contribution");
                    else
                        fprintf(outFile,"room score contribution");
                    break;
                case 0x08:  // Random Number (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    data2(b[x+2], w1);
                    fprintf(outFile," := random[1 to ");
                    if (b[x+6] == 7) {
                        if (w2 < 10)
                            fprintf(outFile,"%d", w2);
                        else
                            fprintf(outFile,"%d (0x%X)", w2, w2);
                    } else {
                        data2(b[x+6], w2);
                    }
                    fprintf(outFile,"] - 1");
                    break;
                case 0x0A:  // Tutorial (unused)
                    break;
                case 0x0B:  // Get Distance To (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    data2(8, w1);   // temp
                    fprintf(outFile," := distance from ");
                    if (b[x+2] & 1)
                        data2(b[x+3], w2);
                    else
                        fprintf(outFile,"Me");
                    fprintf(outFile," to Stack Object");
                    if (b[x+6] & 2)
                        fprintf(outFile," in 1/100ths of a tile");
                    break;
                case 0x0C:  // Get Direction To (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+6]);
                    data2(b[x+2], w1);
                    fprintf(outFile," := direction from ");
                    if (b[x+4] & 1)
                        data2(b[x+5], w2);
                    else
                        fprintf(outFile,"Me");
                    fprintf(outFile," to Stack Object");
                    if (~(b[x+8] & 2))
                        fprintf(outFile," in degrees");
                    break;
                case 0x0D:  // Push Interaction
                    w1 = *(unsigned short *) (&b[x+6]);
                    w2 = *(unsigned short *) (&b[x+9]);
                    w3 = *(unsigned short *) (&b[x+12]);
                    if (b[x+3] & 0x10) {
                        fprintf(outFile,"getting interaction # from ");
                        data2(b[x+5],w1);
                        fprintf(outFile," of ");
                    } else if (b[x+14] & 2)
                        fprintf(outFile,"getting interaction # from results of last FBA call ");
                    else
                        fprintf(outFile,"#0x%X of ", b[x]);
                    if (b[x+3] & 2)
                        data2(0x19, b[x+1]);    // local
                    else
                        data2(9, b[x+1]);       // param
                    fprintf(outFile," onto the stack object's queue, ");
                    fprintf(outFile,"%s", gPriorities[b[x+2]]);
                    if (b[x+3] & 1) {
                        fprintf(outFile,", use icon from ");
                        data2(0x19, b[x+4]);
                    } else if (b[x+14] & 4)
                        fprintf(outFile," use Icon from selector GUID in temp4/5");
                    if (b[x+14] & 8)
                        fprintf(outFile,", Getting Icon Index from Temp6");
                    else
                        fprintf(outFile,", Icon Index is 0x%X", b[x+15]);
                    // if (b[x+3] & 4) fprintf(outFile,", continue as current");
                    if (b[x+3] & 8) fprintf(outFile,", use name");
                    if (b[x+3] & 0x20) fprintf(outFile,", Force run check tree");
                    if (b[x+14] & 1) fprintf(outFile,", Passing first 4 params in");
                    if (b[x+3] & 0x40) {
                        fprintf(outFile,", Linking to interaction with ID in ");
                        data2(b[x+8],w2);
                    }
                    if (b[x+3] & 0x80) {
                        fprintf(outFile,", Returing ID in ");
                        data2(b[x+11],w3);
                    }
                    break;
                case 0x0E:  // Find Best Object for Function
                    w1 = *(unsigned short *) (&b[x+4]);
                    fprintf(outFile,"%s", gFuncTableNames[b[x]]);
                    if (b[x+2] & 1)
                        fprintf(outFile,", Outside Only");
                    else if (b[x+2] & 2)
                        fprintf(outFile,", Inside Only");
                    else if (b[x+2] & 4)
                        fprintf(outFile,", In Room Only");
                    if (b[x+2] & 0x10)
                        fprintf(outFile,", In Line Of Site");
                    if (b[x+2] & 0x40)
                        fprintf(outFile,", Ignoring Lockout");
                    if (b[x+2] & 8) {
                        fprintf(outFile,", Relative to object in ");
                        data2(b[x+3],w1);
                    }

                    break;
                case 0x0F:  // Break Point (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    fprintf(outFile,"if (");
                    data2(b[x+2], w1);
                    fprintf(outFile," != 0)");
                    break;
                case 0x10:  // Find Location For
                    w1 = *(unsigned short *) (&b[x+5]);
                    w2 = *(unsigned short *) (&b[x+8]);
                    if (b[x+2] & 8) {
                        data2(b[x+4], w1);
                        fprintf(outFile," relative to ");
                        data2(b[x+7], w2);
                        switch (b[x]) {
                            case 5:
                                fprintf(outFile,", routing slot in stack var %d",b[x+10]);
                                break;
                            case 6:
                                fprintf(outFile,", routing slot in local var %d",b[x+10]);
                                break;
                            case 7:
                                fprintf(outFile,", literal routing slot %d",b[x+10]);
                                break;
                            case 8:
                                fprintf(outFile,", global routing slot %d",b[x+10]);
                                break;
                        }
                        fprintf(outFile,", facing");
                        if (~(b[x+3] & 1))
                            fprintf(outFile," N");
                        if (~(b[x+3] & 2))
                            fprintf(outFile," NE");
                        if (~(b[x+3] & 4))
                            fprintf(outFile," E");
                        if (~(b[x+3] & 8))
                            fprintf(outFile," SE");
                        if (~(b[x+3] & 0x10))
                            fprintf(outFile," S");
                        if (~(b[x+3] & 0x20))
                            fprintf(outFile," SW");
                        if (~(b[x+3] & 0x40))
                            fprintf(outFile," W");
                        if (~(b[x+3] & 0x80))
                            fprintf(outFile," NW");
                    }
                    else if (b[x+2] & 1) {
                        fprintf(outFile,"start at ");
                        fprintf(outFile,"Stack Object ");       //??
                        data2(0x19,b[x+1]);
                    }
                    if (b[x] != 0) {
                        if (b[x+2] & 8) fprintf(outFile,", ");
                        fprintf(outFile,"%s", gFindGLB[b[x]]);
                    }
//                    if (b[x+2] & 1) {
//                        if (b[x] != 0 || (b[x+2] & 8)) fprintf(outFile,", ");
//                        fprintf(outFile,"start at ");
//                        data2(0x19, b[x+1]);    // local
//                    }
                    if (b[x+2] & 4) fprintf(outFile,", user editable");
                    if ((b[x+2] & 2) == 0) fprintf(outFile,", prefer empty");
                    if (b[x+2] & 0x10) fprintf(outFile,", on level ground");
                    if (b[x+2] & 0x20) fprintf(outFile,", with empty border");
                    if (b[x+2] & 0x40) fprintf(outFile,", begin in front of refobj");
                    if (b[x+2] & 0x80) fprintf(outFile,", with line of site to center");
                    break;
                case 0x11:  // Idle for Input
                    w1 = *(unsigned short *) (&b[x+2]);
                    if (b[x+4] & 1)
                        fprintf(outFile,"Handle Sub Queue Interactions");
                    else {
                        fprintf(outFile,"for ");
                        data2(9, b[x]);
                        fprintf(outFile," ticks, ");
                        if (w1 == 0) fprintf(outFile,"do not ");
                        fprintf(outFile,"allow push");
                    }
                    break;
                case 0x12:  // Remove Object Instance
                    if (b[x] == 0)
                        fprintf(outFile,"Me");
                    else
                        fprintf(outFile,"Stack Object");
                    if (b[x+2] & 1) fprintf(outFile,", return immediately");
                    if (~(b[x+2] & 2)) fprintf(outFile,", cleanup all");
                    break;
                case 0x13:  // Make New Character
                    w1 = *(unsigned short *) (&b[x+4]);
                    w2 = *(unsigned short *) (&b[x+7]);
                    w3 = *(unsigned short *) (&b[x+11]);

                    if (b[x+9] & 8)
                        fprintf(outFile,"getting data from object with GUID in temp0/1");
                    else if (b[x+9] & 0x10)
                        fprintf(outFile,"getting data from object with GUID in temp Token");
                    else if (b[x+9] & 1) {
                        if (b[x+9] & 2)
                            fprintf(outFile,"using Neighbor ID in ");
                        else
                            fprintf(outFile,"using ID in ");
                        data2(b[x+6],w2);
                        if (b[x+9] & 4)
                            fprintf(outFile," as parent 1 and Neighor ID in ");
                        else
                            fprintf(outFile," as parent 1 and ID in ");
                        data2(b[x+3],w1);
                        fprintf(outFile," as parent 2");
                    } else {
                        fprintf(outFile,"age in ");
                        data2(0x19, b[x+1]);
                        fprintf(outFile,", gender in ");
                        data2(0x19, b[x+2]);
                        fprintf(outFile,", skin color in ");
                        data2(0x19, b[x]);
                    }
                    if (b[x+9] & 0x20)
                        fprintf(outFile,", Getting character from Bin"); // ?
                    if (b[x+9] & 0x40) {
                        fprintf(outFile,", Using external guid for Thumbnail Outfit from ");
                        if (b[x+9] & 0x80)
                            fprintf(outFile,"GUID in temp2/3");
                        else
                            data2(b[x+10],w3);
                    }
                    break;
                case 0x14:  // Run Functional Tree
                    fprintf(outFile,"%s", gFuncTableNames[b[x]]);
                    c1 = b[x+2];
                    if (c1 & 1) fprintf(outFile,", change icon");
                    if (c1 & 2) fprintf(outFile,", passing parameters from calling tree");
                    if (c1 & 4) fprintf(outFile,", running check tree only");
                    break;
                case 0x16:  // Turn Body Towards
                    fprintf(outFile,"%s", gTurnBody[b[x]]);
                    break;
                case 0x17:  // Play / Stop Sound Event (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    if (b[x+4] & 0x40)
                        fprintf(outFile,"Stop Sound ");
                    else
                        fprintf(outFile,"Play Sound ");
                    if (w1 < 10000)
                        fprintf(outFile,"Private id = 0x%X", w1);
                    else if (w1 < 20000)
                        fprintf(outFile,"Global id = 0x%X", w1-10000);
                    else
                        fprintf(outFile,"SemiGlobal id = 0x%X", w1-20000);
                    if (b[x+4] & 2) fprintf(outFile,", with stack obj as source");
                    if (~(b[x+4] & 2)) {
                        if (b[x+4] & 0x10)
                            fprintf(outFile,", auto vary");
                        if (w2 != 0)
                            fprintf(outFile,", sampled at %d, w2"); // ?
                        if (b[x+5] != 0)
                            fprintf(outFile,", volume %d", b[x+5]); // ?
                    }
                    break;
                case 0x19:  // Alter Budget
                    c1 = b[x+1];
                    switch (b[x]) {
                        case 0:
                            c1 = 7;     // literal
                            break;
                        case 1:
                            c1 = 9;     // param
                            break;
                        case 2:
                            c1 = 0x19;  // local
                            break;
                    }
                    w1 = *(unsigned short *) (&b[x+2]);
                    w2 = *(unsigned short *) (&b[x+7]);
                    switch (b[x+4] & 3) {
                        case 0:
                            fprintf(outFile,"subtract ");
                            data2(c1, w1);
                            break;
                        case 1:
                            fprintf(outFile,"test if ");
                            data2(c1, w1);
                            fprintf(outFile," may be subtracted");
                            break;
                        case 2:
                            fprintf(outFile,"add ");
                            data2(c1, w1);
                            break;
                        case 3:
                            fprintf(outFile,"test if ");
                            data2(c1, w1);
                            fprintf(outFile," may be added");
                            break;
                    }
                    if (b[x+4] & 4)
                        fprintf(outFile," Multiplied by value in Temp2");
                    else
                        fprintf(outFile," Multiplied by 0x%X", w2);
                    fprintf(outFile," as %s", gExpenses[b[x+6]]);
                    break;
                case 0x1A:  // Relationship
                    c1 = b[x];      // var
                    c2 = b[x+1];    // flags
                    w2 = *(unsigned short *) (&b[x+6]);

                    if (f < 0x8007) {   // old-style parameter usage
                        if ((c2 & 4) == 0) {
                            data2(b[x+4], w2);
                            fprintf(outFile," := ");
                        }
                        fprintf(outFile,"var %d (%s) of ", c1, gRelLabels[c1]);
                        switch (c2 & 3) {
                            case 0:
                                fprintf(outFile,"Me to Stack Object");
                                break;
                            case 1:
                                fprintf(outFile,"Stack Object to Me");
                                break;
                            case 2:
                                fprintf(outFile,"Stack Object to ");
                                data2(0x19, b[x+3]);
                                break;
                            case 3:
                                data2(0x19, b[x+3]);
                                fprintf(outFile," to Stack Object");
                                break;
                        }
                        if (c2 & 4) {
                            fprintf(outFile," := ");
                            data2(b[x+4], w2);
                        }
                    } else {            // new-style parameter usage
                        w1 = *(unsigned short *) (&b[x+9]);
                        w3 = *(unsigned short *) (&b[x+3]);
                        fprintf(outFile,"Access var %d (%s) of ", c1, gRelLabels[c1]);
                        data2(b[x+2],w3);
                        fprintf(outFile," to ");
                        data2(b[x+5],w2);
                        fprintf(outFile,". ");
                        if (c2 & 4)
                            fprintf(outFile,"Get value from ");
                        else
                            fprintf(outFile,"Put value into ");
                        data2(b[x+8], w1);
                    }
                    if (c2 & 1) fprintf(outFile,", fail if too small");
                    if (c2 & 2) fprintf(outFile,", use neighbors");
                    break;
                case 0x1B:  // Go To Relative Position
                    c1 = (b[x+2] + 2) & 0xFF;
                    c2 = (b[x+3] + 2) & 0xFF;
                    fprintf(outFile,"Location = %s, ", gRelLocations[c1]);
                    fprintf(outFile,"Direction = %s", gRelDirections[c2]);
                    c3 = b[x+6];
                    if (c3 & 2) fprintf(outFile,", no failure trees");
                    if (c3 & 4) fprintf(outFile,", allow different altitudes");
                    break;
                case 0x1C:  // Run Tree by Name
                    c1 = b[x+4] - 1;
                    w1 = *(unsigned short *) (&b[x+7]);
                    w2 = *(unsigned short *) (&b[x+10]);
                    w3 = *(unsigned short *) (&b[x+13]);
                    if (b[x+2] & 1)
                        fprintf(outFile,"%s", gNamedTreePrim[c1]);
                    else if (b[x+2] & 2) {
//                        fprintf(outFile,".", c1);          // ?
//                        if (gGlobGroup) {
                            if (readString2(gGlobGroup, 0x12F, c1) == 0)
                                fprintf(outFile,"[SemiGlobal STR# 0x12F:0x%X]", c1);
//                        } else
//                            if (readString2(gGroup, 0x12F, c1) == 0)
//                                fprintf(outFile,"[SemiGlobal STR# 0x12F:0x%X]", c1);
                    } else {
//                        fprintf(outFile,".", c1);
                        if (readString2(gGroup, 0x12F, c1) == 0)
                            fprintf(outFile,"[STR# 0x12F:0x%X]", c1);
                    }
                    if (b[x+2] & 0x20)
                        fprintf(outFile,", Passing in params from the current tree");
                    else if (b[x+2] & 0x10) {
                        fprintf(outFile,", Passing in params where param 0 = ");
                        data2(b[x+6], w1);
                        fprintf(outFile,", param 1 = ");
                        data2(b[x+9], w2);
                        fprintf(outFile,", param 2 = ");
                        data2(b[x+12], w3);
                    }
                    switch (b[x+5]) {
                        case 0:
                            fprintf(outFile,", run in my stack");
                            break;
                        case 1:
                            fprintf(outFile,", run in Stack Object's stack");
                            break;
                        case 2:
                            fprintf(outFile,", push onto my stack");
                            break;
                    }
                    break;
                case 0x1D:  // Set Motive Change (false = error)
                    w1 = *(unsigned short *) (&b[x+4]);
                    w2 = *(unsigned short *) (&b[x+6]);
                    if (b[x+3] & 1)
                        fprintf(outFile,"clear all");
                    else {
                        data2(0xE, b[x+2]);     // my motives
                        fprintf(outFile," += ");
                        data2(b[x], w1);
                        fprintf(outFile," per hour, stop at ");
                        data2(b[x+1], w2);
                    }
                    if (b[x+3] & 2)
                        fprintf(outFile,", Auto Clearing the Person Data Motive Decay value");
                    break;
                case 0x1E:  // Gosub Found Action
                    switch (b[x]) {
                        case 0:
                            fprintf(outFile,"Gosub Found Action");
                            break;
                        case 1:
                            fprintf(outFile,"Gosub Interrupt Action");
                            break;
                        case 2:
                            fprintf(outFile,"Gosub Reaction Action");
                            break;
                        case 3:
                            fprintf(outFile,"Gosub Top User Action");
                            break;
                    }
                    break;

                case 0x1F:  // Set to Next (false = next not found)
                    if (b[x+5] != 0xA || b[x+7] != 0 && (b[x+4] & 0x80) != 0) {
                        data2(b[x+5], b[x+7]);
                        fprintf(outFile," := next ");
                    }
                    c1 = (b[x+4] & 0x7F);
                    fprintf(outFile,"%s", gNextObj[c1]);
                    d1 = *(unsigned long *) (&b[x]);
                    if (c1 == 4 || c1 == 7) fprintf(outFile," GUID 0x%08X", d1);
                    readGUID(d1);
                    if (c1 == 9 || c1 == 0x22) data2(0x19, b[x+6]);   // local
                    break;
                case 0x20:  // Test Object Type
                    w1 = *(unsigned short *) (&b[x+4]);
                    d1 = *(unsigned long *) (&b[x]);
                    if (b[x+7] & 4)
                        fprintf(outFile,"Return GUID selected in Temp0/1 ");
                    else {
                        fprintf(outFile,"type of ");
                        data2(b[x+6], w1);
                    }
                    if (d1 == 0x4C7CAB2B)
                        fprintf(outFile,"the GUID of the temporary inventory token");
                    else {
                        fprintf(outFile," == GUID 0x%08X", d1);
                        readGUID(d1);
                    }
                    if (~(b[x+7] & 4)) {
                        if (b[x+7] & 1)
                            fprintf(outFile," Checking against original, not current GUID");
                        if (b[x+7] & 2)
                            fprintf(outFile,", incoming ID is a neighbor ID");
                    }



                    break;
                case 0x21:  // Find 5 Worst Motives (false = error)
                    w1 = *(unsigned short *) (&b[x+4]);
                    w2 = *(unsigned short *) (&b[x+6]);
                    switch (w1) {
                        case 0:
                            fprintf(outFile,"My");
                            break;
                        case 1:
                            fprintf(outFile,"Stack Object's");
                            break;
                        case 2:
                            fprintf(outFile,"Target Object's");
                            break;
                    }
                    fprintf(outFile," lowest ");
                    switch (w2) {
                        case 0:
                            fprintf(outFile,"Motive");
                            break;
                        case 1:
                            fprintf(outFile,"Mental Motive");
                            break;
                        case 2:
                            fprintf(outFile,"Physical Motive");
                            break;
                    }
                    fprintf(outFile," into temps 0-4");
                    break;
                case 0x22:  // UI Effect (false = error)
                    c1 = b[x];
                    c2 = b[x+5]; // flags
                    w1 = *(unsigned short *) (&b[x+1]);
                    w2 = *(unsigned short *) (&b[x+3]);
                    w3 = *(unsigned short *) (&b[x+6]);
                    w4 = *(unsigned short *) (&b[x+8]);
                    w5 = *(unsigned short *) (&b[x+11]);

                    switch (c1) {
                        case 0:
                            fprintf(outFile,"Press Control ");
                            break;
                        case 1:
                            fprintf(outFile,"Disable Control ");
                            break;
                        case 2:
                            fprintf(outFile,"Enable Control ");
                            break;
                        case 3:
                            fprintf(outFile,"Play Effect on Control ");
                            break;
                        case 4:
                            fprintf(outFile,"Set Event tree on Control ");
                            break;
                        case 5:
                            fprintf(outFile,"Reset State on All Controls ");
                            break;
                        case 6:
                            fprintf(outFile,"Disable All Controls ");
                            break;
                        case 7:
                            fprintf(outFile,"Reset event trees on All Controls ");
                            break;
                        case 8:
                            fprintf(outFile,"Set Event Tree on TNS node ");
                            break;
                        case 9:
                            fprintf(outFile,"Set Control Visible ");
                            break;
                        case 0xA:
                            fprintf(outFile,"Set Control Hidden ");
                            break;
                    }
                    if (c1 < 5 || c1 > 8)
                        if (c2 & 4) {
                            if (readString2(0x7FD46CD0, 0x96, w2) == 0)
                                fprintf(outFile,"[Global STR# 0x96:0x%X]", w2);
                        }
                        else if (c2 & 8){
                            if (readString2(gGlobGroup, 0x96, w2) == 0)
                                fprintf(outFile,"[SemiGlobal STR# 0x96:0x%X]", w2);
                        }
                        else {
                            if (readString2(gGroup, 0x96, w2) == 0)
                                fprintf(outFile,"[Private STR# 0x96:0x%X]", w2);
                        }
                    if (c1 != 8) {
                        fprintf(outFile," in window ");
                        if (c2 & 1) {
                            if (readString2(0x7FD46CD0, 0x96, w1) == 0)
                                fprintf(outFile,"[Global STR# 0x96:0x%X]", w1);
                        }
                        else if (c2 & 2){
                            if (readString2(gGlobGroup, 0x96, w1) == 0)
                                fprintf(outFile,"[SemiGlobal STR# 0x96:0x%X]", w1);
                        }
                        else {
                            if (readString2(gGroup, 0x96, w1) == 0)
                                fprintf(outFile,"[Private STR# 0x96:0x%X]", w1);
                        }
                    }
                    if (c1 == 3)
                        if (w3 != 0)
                            fprintf(outFile," Starting Effect ");
                        else
                            fprintf(outFile," Stopping Effect ");
                    if (c1 == 4 || c1 == 8)
                        if (w4 == 0)
                            fprintf(outFile," No Event Tree");
                        else {
                            switch (b[x+10]) {
                                case 0:
                                    fprintf(outFile," Using Private Event Tree: ");
                                    break;
                                case 1:
                                    fprintf(outFile," Using SemiGlobal Event Tree: ");
                                    break;
                                default:
                                    fprintf(outFile," Using Global Event Tree: ");
                                    break;
                            }
                            fprintf(outFile,"0x%X",w4);
                            fprintf(outFile," (");
                            readFn2(w4);
                            fprintf(outFile,")");

                        }
                    if (c1 == 8) {
                        fprintf(outFile,", Getting TNS ID from ");
                        data2(b[x+13],w5);
                    }
                    break;
                case 0x23:  // Camera Control (false = error)
                    c1 = b[x+4];    // flags
                    w1 = *(unsigned short *) (&b[x]);
                    if (c1 & 1) {
                        fprintf(outFile,"show stack obj");
                        switch (b[x+3]) {
                            case 1:
                                fprintf(outFile,", far");
                                break;
                            case 2:
                                fprintf(outFile,", mid");
                                break;
                            case 3:
                                fprintf(outFile,", near");
                                break;
                        }
                        fprintf(outFile," zoom");
                        if (~(c1 & 0x40))
                            fprintf(outFile,", slow down");
                        if (c1 & 8)
                            fprintf(outFile,", center");
                        if (c1 & 0x20)
                            fprintf(outFile,", using timeout in temp 0");
                        else {
                            fprintf(outFile,", using timeout of ");
                            if (w1 < 10)
                                fprintf(outFile,"%d", w1);
                            else
                                fprintf(outFile,"%d (0x%X)", w1, w1);
                        }
                    } else
                        fprintf(outFile,"un-show stack obj");
                    break;
/*
                case 0x24:  // Dialog
                    c1 = b[x+5];    // prim
                    c2 = (b[x+7] >> 4) & 7; // temp
                    c3 = b[x+8]; // flags
                    w1 = *(unsigned short *) (&b[x+13]); // message
                    w2 = b[x+2] << 8 + b[x]; // cancel
                    w3 = b[x];
                    w4 = b[x+2];

                    switch (c1) {
                        case 9:
                            fprintf(outFile,"Text Notification Stop ");
                            fprintf(outFile,", Getting Text Notification ID from temp %d", c2);
                            switch (b[x+12]) {
                                case 0:
                                    fprintf(outFile,", TNS Sim Type");
                                    break;
                                case 1:
                                    fprintf(outFile,", TNS System Type");
                                    break;
                            }
                            break;
                        case 0xB:
                            fprintf(outFile,"Magazine Rack ");
                            break;
                        case 0xC:
                            fprintf(outFile,"Jewelry Rack ");
                            break;
                        case 0xD:
                            fprintf(outFile,"VideoGame Selection ");
                            break;
                        case 0xE:
                            fprintf(outFile,"Food Selection ");
                            fprintf(outFile,"passing through data starting with local var %d", b[x+11]);
                            break;
                        case 0x10:
                            fprintf(outFile,"Clothing Purchase ");
                            break;
                        case 0x11:
                            fprintf(outFile,"Clothing Selection ");
                            break;
                        case 0x12:
                            fprintf(outFile,"Tutorial Tasks ");
                            break;
                        case 0x14:
                            fprintf(outFile,"Clothing Try On ");
                            break;
                        default:
                            switch (c1) {
                                case 0:
                                    fprintf(outFile,"message: ");
                                    break;
                                case 1:
                                    fprintf(outFile,"choice: ");
                                    break;
                                case 2:
                                    data2(8, c2); // Temp
                                    fprintf(outFile," := tri choice: ");
                                    break;
                                case 3:
                                    fprintf(outFile,"text entry: ");
                                    break;
                                case 4:
                                    fprintf(outFile,"tutorial: ");
                                    break;
                                case 5:
                                    fprintf(outFile,"Phone Services ");
                                    break;
                                case 6:
                                    fprintf(outFile,"Phone Invite ");
                                    break;
                                case 7:
                                    fprintf(outFile,"Phone Party ");
                                    break;
                                case 8:
                                    fprintf(outFile,"Text Notification ");
                                    break;
                                case 0xA:
                                    fprintf(outFile,"Text Notification Modify ");
                                    break;
                                case 0xF:
                                    fprintf(outFile,"Phone Special Services ");
                                    break;
                                case 0x13:
                                    fprintf(outFile,"ListSelect Dialog ");
                                    break;
                                case 0x16:
                                    fprintf(outFile,"Tutorial Next ");
                                    break;
                                case 0x17:
                                    fprintf(outFile,"Set Baby Name ");
                                    break;
                                case 0x18:
                                    fprintf(outFile,"Aspirations ");
                                    break;
                                case 0x19:
                                    fprintf(outFile,"Tutorial Next Modal");
                                    break;
                            }
                            if (c3 & 2) // ?
                                fprintf(outFile,", expecting message Index in temp 0");
                            else if (c3 & 1) {
                                if (readString2(gGlobGroup, 0x12D, w1 - 1) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x12D:0x%X]", w1 - 1);
                                fprintf(outFile, "---");
                                if (readString2(gGlobGroup, 0x12D, w3 - 1) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x12D:0x%X]", w3 - 1);
                            }
                            else if (c3 & 0x40) {
                                fprintf(outFile,"%s", gDialogPrim[w1 - 1]);
                                fprintf(outFile, "---");
                                fprintf(outFile,"%s", gDialogPrim[w3 - 1]);
                            }
                            else {
                                if (readString2(gGroup, 0x12D, w1 - 1) == 0)
                                    fprintf(outFile,"[Private STR# 0x12D:0x%X]", w1 - 1);
                                fprintf(outFile, "---");
                                if (readString2(gGroup, 0x12D, w3 - 1) == 0)
                                    fprintf(outFile,"[Private STR# 0x12D:0x%X]", w3 - 1);
                            }
                            if (c3 & 0x20) // ?
                                fprintf(outFile,", expecting Cancel Index in temp 0");
                            else {
                                fprintf(outFile,", Cancel: ");
                                if (w2 != 0)
                                    if (readString2(gGroup, 0x12D, w2 - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", w2 - 1);
                                fprintf(outFile, "---");
                                if (w4 != 0)
                                    if (readString2(gGroup, 0x12D, w4 - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", w2 - 1);
                            }
                            break;
                    }
                    break;

*/

                case 0x24:  // Dialog
                    c1 = b[x+5];    // prim
                    c2 = (b[x+7] >> 4) & 7; // temp
                    c3 = b[x+8]; // flags
                    if (nodeVersion) {
                        w1 = *(unsigned short *) (&b[x+13]); // message
                        w2 = b[x+2] << 8 + b[x]; // cancel
                    } else {
                        w1 = b[x+2]; // message
                        w2 = b[x]; // cancel
                    }
                    switch (c1) {
                        case 9:
                            fprintf(outFile,"Text Notification Stop ");
                            fprintf(outFile,", Getting Text Notification ID from temp %d", c2);
                            switch (b[x+12]) {
                                case 0:
                                    fprintf(outFile,", TNS Sim Type");
                                    break;
                                case 1:
                                    fprintf(outFile,", TNS System Type");
                                    break;
                            }
                            break;
                        case 0xB:
                            fprintf(outFile,"Magazine Rack ");
                            break;
                        case 0xC:
                            fprintf(outFile,"Jewelry Rack ");
                            break;
                        case 0xD:
                            fprintf(outFile,"VideoGame Selection ");
                            break;
                        case 0xE:
                            fprintf(outFile,"Food Selection ");
                            fprintf(outFile,"passing through data starting with local var %d", b[x+11]);
                            break;
                        case 0x10:
                            fprintf(outFile,"Clothing Purchase ");
                            break;
                        case 0x11:
                            fprintf(outFile,"Clothing Selection ");
                            break;
                        case 0x12:
                            fprintf(outFile,"Tutorial Tasks ");
                            break;
                        case 0x14:
                            fprintf(outFile,"Clothing Try On ");
                            break;
                        default:
                            switch (c1) {
                                case 0:
                                    fprintf(outFile,"message: ");
                                    break;
                                case 1:
                                    fprintf(outFile,"choice: ");
                                    break;
                                case 2:
                                    data2(8, c2); // Temp
                                    fprintf(outFile," := tri choice: ");
                                    break;
                                case 3:
                                    fprintf(outFile,"text entry: ");
                                    break;
                                case 4:
                                    fprintf(outFile,"tutorial: ");
                                    break;
                                case 5:
                                    fprintf(outFile,"Phone Services ");
                                    break;
                                case 6:
                                    fprintf(outFile,"Phone Invite ");
                                    break;
                                case 7:
                                    fprintf(outFile,"Phone Party ");
                                    break;
                                case 8:
                                    fprintf(outFile,"Text Notification ");
                                    break;
                                case 0xA:
                                    fprintf(outFile,"Text Notification Modify ");
                                    break;
                                case 0xF:
                                    fprintf(outFile,"Phone Special Services ");
                                    break;
                                case 0x13:
                                    fprintf(outFile,"ListSelect Dialog ");
                                    break;
                                case 0x16:
                                    fprintf(outFile,"Tutorial Next ");
                                    break;
                                case 0x17:
                                    fprintf(outFile,"Set Baby Name ");
                                    break;
                                case 0x18:
                                    fprintf(outFile,"Aspirations ");
                                    break;
                                case 0x19:
                                    fprintf(outFile,"Tutorial Next Modal");
                                    break;
                            }
                            if (c3 & 2) // ?
                                fprintf(outFile,", expecting message Index in temp 0");
                            else if (c3 & 1) {
                                if (readString2(gGlobGroup, 0x12D, w1 - 1) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x12D:0x%X]", w1 - 1);
                            }
                            else if (c3 & 0x40)
                                fprintf(outFile,"%s", gDialogPrim[w1 - 1]);
                            else if (readString2(gGroup, 0x12D, w1 - 1) == 0)
                                fprintf(outFile,"[Private STR# 0x12D:0x%X]", w1 - 1);
                            if (c1 == 8 || c1 == 0xA) {
                                switch (b[12]) {
                                    case 0:
                                        fprintf(outFile,", TNS Sim Type");
                                        break;
                                    case 1:
                                        fprintf(outFile,", TNS System Type");
                                        break;
                                    case 2:
                                        fprintf(outFile,", TNS System Dialog Type");
                                        break;
                                    case 3:
                                        fprintf(outFile,", TNS Birthday Type");
                                        break;
                                }
                                fprintf(outFile,", priority %d", b[x+9] + 1);
                                fprintf(outFile,", timeout %d", b[x+10]);
                                if (c1 == 0xA)
                                    fprintf(outFile,", getting Text ID in temp ");
                                else
                                    fprintf(outFile,", putting Text ID in temp ");
                                fprintf(outFile,"%d",c2);
                            }
                            else {
                                if (c3 & 4) // ?
                                    fprintf(outFile,", expecting Yes button Index in temp 0");
                                else if (b[x+3] != 0) {
                                    fprintf(outFile,", Yes: ");
                                    if (readString2(gGroup, 0x12D, b[x+3] - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", b[x+3] - 1);
                                }
                                if (c3 & 8) // ?
                                    fprintf(outFile,", expecting No button Index in temp 0");
                                else if (b[x+4] != 0) {
                                    fprintf(outFile,", No: ");
                                    if (readString2(gGroup, 0x12D, b[x+4] - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", b[x+4] - 1);
                                }
                                if (c3 & 0x10) // ?
                                    fprintf(outFile,", expecting Title Index in temp 0");
                                else if (b[x+6] != 0) {
                                    fprintf(outFile,", Title: ");
                                    if (readString2(gGroup, 0x12D, b[x+6] - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", b[x+6] - 1);
                                }
                                if (c3 & 0x20) // ?
                                    fprintf(outFile,", expecting Cancel Index in temp 0");
                                else if (w2 != 0) {
                                    fprintf(outFile,", Cancel: ");
                                    if (readString2(gGroup, 0x12D, w2 - 1) == 0)
                                        fprintf(outFile,"[STR# 0x12D:0x%X]", w2 - 1);
                                }
                                if ((b[x+7] & 0x81) != 0) {
                                    if (b[x+7] & 1)
                                        fprintf(outFile," return");
                                    else
                                        fprintf(outFile," engage");
                                    fprintf(outFile," and ");
                                    if (c1 & 0x80)
                                        fprintf(outFile,"continue sim");
                                    else
                                        fprintf(outFile,"block sim");
                                }
                            }
                            if (w1 != 0x16 && w1 != 0x19 && (b[x+7] & 0xE) != 2) {
                                fprintf(outFile,", icon: ");
                                switch ((b[x+7] >> 1)& 7) {
                                    case 3:
                                        fprintf(outFile,"BMP = 0x%X", (b[x+1] + 5000));
                                        break;
                                    case 4:
                                        if (readString2(gGroup, 0x12D, b[x+1] - 1) == 0)
                                            fprintf(outFile,"[STR# 0x12D:0x%X]", b[x+1] - 1);
                                        break;
                                    default:
                                        fprintf(outFile,"%s", gIconTypes[(b[x+7] >> 1) & 7]);
                                        break;
                                }
                            }
                            break;
                    }
                    break;

                case 0x25:  // Test Sim Interacting With
                    fprintf(outFile,"me with stack obj.");
                    break;
                case 0x2A:  // Create New Object Instance
                    if (b[x+5] & 4)
                        fprintf(outFile,"neighbor in stack obj");
                    else if (b[x+5] & 0x40)
                        fprintf(outFile,"from GUID in temp Inventory Token");
                    else if (b[x+5] & 0x80)
                        fprintf(outFile,"from GUID in temp0/1");
                    else {
                        d1 = *(unsigned long *) (&b[x]);
                        if (d1 == 0)
                            fprintf(outFile,"GUID 0");
                        else {
                            fprintf(outFile,"GUID 0x%08X", d1);
                            readGUID(d1);
                        }
                    }
                    fprintf(outFile,", place ");
                    c1 = b[x+4];
                    switch (c1) {
                        case 4:
                            fprintf(outFile,"in stack objs slot %d", b[x+9]);
                            break;
                        case 0xA:
                            fprintf(outFile,"in obj in temp 0's slot %d", b[x+9]);
                            break;
                        default:
                            fprintf(outFile,"%s", gObjPlace[c1]);
                            break;
                    }
                    if (c1 == 8 || c1 == 9) {
                        fprintf(outFile," ");
                        data2(0x19, b[x+6]);     // local
                    }
                    c2 = b[x+5];
                    if (c2 & 2)
                        fprintf(outFile,", pass object ids to main");
                    else if (c2 & 0x10)
                        fprintf(outFile,", pass temp 0 to main");
                    if (c2 & 1)
                        fprintf(outFile,", do not duplicate");
                    if (c2 & 8)
                        fprintf(outFile,", fail if tile is non-empty");
                    if (b[x+10] & 1)
                        fprintf(outFile,", Moving in a new Sim");
                    if (b[x+10] & 2)
                        fprintf(outFile,", copying design mode materials from object in temp5");
                    break;
                case 0x2D:  // Go To Routing Slot
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    switch (w2) {
                        case 0:
                            fprintf(outFile,"in ");
                            data2(9, w1);   // param
                            break;
                        case 1:
                            if (b[x+4] & 2)
                                fprintf(outFile,"From Slot index in Temp 0", w1);
                            else
                                fprintf(outFile,"0x%X", w1);
                            break;
                        case 2:
                            fprintf(outFile,"global 0x%X", w1);
                            break;
                        default:
                            fprintf(outFile,"??? 0x%X", w1);
                    }
                    if (b[x+4] & 1) fprintf(outFile,", no failure trees");
                    if (b[x+4] & 4) fprintf(outFile,", ignoring dest obj footprint");
                    if (b[x+4] & 8) fprintf(outFile,", allow different altitudes");
                    break;
                case 0x2E:  // Snap
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    w3 = *(unsigned short *) (&b[x+8]);
                    fprintf(outFile,"%s", gSnapMode[w2]);
                    if (w2 == 0 || w2 == 3 || w2 == 4)
                        if (b[x+4] & 8)
                            fprintf(outFile," to slot in temp 0");
                        else
                            fprintf(outFile," 0x%X", w1);
                    if (w3 == 1)
                        fprintf(outFile,", from slot in temp 1");
                    if (b[x+4] & 2)
                        fprintf(outFile,", ask person to move");
                    if (b[x+4] & 0x10)
                        fprintf(outFile,", TEST ONLY");
                    break;
                case 0x30:  // Stop ALL Sounds (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    if (w1 == 0)
                        fprintf(outFile,"of Me");
                    else
                        fprintf(outFile,"of Stack Object");
                    break;
                case 0x31:  // Notify the Stack Object out of Idle (implied, false = error)
                    break;
                case 0x32:  // Add/Change the Action String (false = error)
                    c2 = b[x+2]; // flags
                    if (b[x+9] == 0) {
                        c1 = b[x+4] - 1;
                        fprintf(outFile,"Add / Change Interaction string Mode, ");
                        if (c2 & 0x10) {
                            fprintf(outFile,"Getting index from temp 0");
                            if (c2 & 4)
                                fprintf(outFile,", getting string from globals");
                            else if (c2 & 8)
                                fprintf(outFile,", getting string from semiglobals");
                            else
                                fprintf(outFile,", getting string from privates");
                        }
                        else if (c2 & 4)
                            fprintf(outFile,"%s", gActionString[c1]);
                        else if (c2 & 8) {
                            if (readString2(gGlobGroup, 0x12E, c1) == 0)
                                fprintf(outFile,"[SemiGlobal STR# 0x12E:0x%X]", c1);
                        }
                        else if (readString2(gGroup, 0x12E, c1) == 0)
                            fprintf(outFile,"[Private STR# 0x12E:0x%X]", c1);
                    } else {
                        fprintf(outFile,"Interaction Icon Change Mode");
                        if (c2 & 0x20) {
                            fprintf(outFile,", Using external guid for Thumbnail Outfit from ");
                            if (c2 & 0x40)
                                fprintf(outFile,"GUID in temp2/3");
                            else {
                                d1 = *(unsigned long *) (&b[x+5]);
                                if (d1 == 0)
                                    fprintf(outFile,"GUID 0");
                                else {
                                   fprintf(outFile,"GUID 0x%08X", d1);
                                   readGUID(d1);
                                }
                            }
                        } else {
                            w1 = *(unsigned short *) (&b[x+12]);
                            fprintf(outFile,", Using object ID in ");
                            data2(b[x+11], w1);
                        }
                        if (c2 & 0x80)
                            fprintf(outFile,", Getting icon index into model table from temp 1");
                        else
                            fprintf(outFile,", Using index of %d",b[x+10]);
                    }
                    break;
                case 0x33:  // Manage Inventory
                    w1 = *(unsigned short *) (&b[x+14]);
                    w2 = *(unsigned short *) (&b[x+11]);
                    w3 = *(unsigned short *) (&b[x+2]);
                    fprintf(outFile,"Access the ");
                    switch (b[x] & 3) {
                        case 0:
                            fprintf(outFile,"Global ");
                            break;
                        case 1:
                            fprintf(outFile,"Lot ");
                            break;
                        case 2:
                            fprintf(outFile,"Family ");
                            break;
                        case 3:
                            fprintf(outFile,"Neighbor ");
                            break;
                    }
                    if (b[x] & 4)
                        fprintf(outFile,"Counted Inventory");
                    else
                        fprintf(outFile,"Singular Inventory");
                    if ((b[x] & 3) != 0) {
                        fprintf(outFile," from ID ");
                        data2(b[x+1], w3);
                    }
                    fprintf(outFile,". with category %d",b[x+9]);
                    d1 = *(unsigned long *) (&b[x+5]);
                    if (d1 != 0) {
                        fprintf(outFile," GUID 0x%08X", d1);
                        readGUID(d1);
                        fprintf(outFile,". "); //?
                    }
                    else
                        fprintf(outFile," of GUID from Stack Object. ");
                    if (b[x] & 4)
                        switch (b[x+4]) {
                            case 0:
                                fprintf(outFile,"Add token");
                                fprintf(outFile,". Pull count from ");
                                data2(b[x+13], w1);
                                break;
                            case 1:
                                fprintf(outFile,"Add to token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,". Pull count from ");
                                data2(b[x+13], w1);
                                break;
                            case 2:
                                fprintf(outFile,"Remove token");
                                fprintf(outFile,". Pull count from ");
                                data2(b[x+13], w1);
                                break;
                            case 3:
                                fprintf(outFile,"Remove to token at index from");
                                data2(b[x+10], w2);
                                fprintf(outFile,". Pull count from ");
                                data2(b[x+13], w1);
                                break;
                            case 4:
                                fprintf(outFile,"Remove all tokens");
                                fprintf(outFile,".");
                                break;
                            case 5:
                                fprintf(outFile,"Remove all tokens from token at index from");
                                data2(b[x+10], w2);
                                fprintf(outFile,".");
                                break;
                            case 6:
                                fprintf(outFile,"Find the token");
                                fprintf(outFile,". Put count into ");
                                data2(b[x+13], w1);
                                fprintf(outFile,".");
                                break;
                            case 7:
                                fprintf(outFile,"Read token");
                                fprintf(outFile," into My Temp Token.");
                                break;
                            case 8:
                                fprintf(outFile,"Read token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile," into My Temp Token.");
                                break;
                            case 9:
                                fprintf(outFile,"Set To Next token");
                                fprintf(outFile,". Starting at index from ");
                                data2(b[x+10], w2);
                                break;
                            case 0xA:
                                fprintf(outFile,"Store the count of the tokens in this inventory into ");
                                data2(b[x+13], w1);
                                fprintf(outFile,".");
                                break;
                        }
                    else
                        switch (b[x+4]) {
                            case 0:
                                fprintf(outFile,"Add token");
                                fprintf(outFile,".");
                                break;
                            case 1:
                                fprintf(outFile,"Remove token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,".");
                                break;
                            case 2:
                                fprintf(outFile,"Remove at tokens");
                                fprintf(outFile,".");
                                break;
                            case 3:
                                fprintf(outFile,"Set To Next token");
                                fprintf(outFile,". Starting at index from ");
                                data2(b[x+10], w2);
                                break;
                            case 4:
                                fprintf(outFile,"Push property on token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,". Get property value from ");
                                data2(b[x+13], w1);
                                break;
                            case 5:
                                fprintf(outFile,"Pop property off token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,". Put property value into ");
                                data2(b[x+13], w1);
                                fprintf(outFile,".");
                                break;
                            case 6:
                                fprintf(outFile,"Read token into My Temp Token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,".");
                                break;
                            case 7:
                                fprintf(outFile,"Get property from token in My Temp Token at index from ");
                                data2(b[x+10], w2);
                                fprintf(outFile,". Put property value into ");
                                data2(b[x+13], w1);
                                fprintf(outFile,".");
                                break;
                            case 8:
                                break;
                            case 9:
                                fprintf(outFile,"Save My Temp Token back to the location it was loaded from.");
                                break;
                            case 0xA:
                                fprintf(outFile,"Store the count of the tokens in this inventory into ");
                                data2(b[x+13], w1);
                                fprintf(outFile,".");
                                break;
                        }
                    break;
                case 0x69:  // Animate Object (false = error)
                    c1 = b[x+2];
                    c2 = b[x+10];
                    w1 = *(unsigned short *) (&b[x+7]);
                    w2 = *(unsigned short *) (&b[x]);
                    w3 = *(unsigned short *) (&b[x+4]);
                    if (c1 & 4)
                        fprintf(outFile,"anim id in STR# 0x86:[Param %d]", w2);
                    else {
                        if (readString2(gGroup, 0x86, w2) == 0)
                            fprintf(outFile,"anim id in STR# 0x86:%X", w2);
                    }
                    fprintf(outFile," affecting ");
                    data2(b[x+6],w1);       // target object
                    if (w3 == 0)
                        fprintf(outFile,", No Event Tree");
                    else {
                        switch (b[x+9]) {
                            case 0:
                                fprintf(outFile,", Private Event Tree: ");
                                break;
                            case 1:
                                fprintf(outFile,", SemiGlobal Event Tree: ");
                                break;
                            default:
                                fprintf(outFile,", Global Event Tree: ");
                                break;
                        }
                        fprintf(outFile,"0x%X",w3);
                        fprintf(outFile," (");
                        readFn2(w3);
                        fprintf(outFile,")");

                    }
                    if (c1 & 8)
                        fprintf(outFile,", Interruptible");
                    if (c1 & 1)
                        fprintf(outFile,", Flipped");
                    if (c1 & 0x10)
                        fprintf(outFile,", Start at tag in Temp 0");
                    if (c1 & 0x20)
                        fprintf(outFile,", Loop Count in Temp 1");
                    if (c1 & 0x80)
                        fprintf(outFile,", No Blend in");
                    if (c1 & 0x40)
                        fprintf(outFile,", No Blend out");
                    if (c1 & 2)
                        fprintf(outFile,", Anim Speed in Temp2");
                    if (c2 & 1)
                        fprintf(outFile,", Flip Flag in Temp 3");
                    if (c2 & 4)
                        fprintf(outFile,", Sync to calling objects Anim");
                    if (c2 & 8)
                        fprintf(outFile,", Align blend out with to calling objects Anim");
                    if (c2 & 0x80)
                        fprintf(outFile,", Not Hurryable");
                    break;
                case 0x6A:  // Animate Sim
                    c1 = b[x+2];
                    c2 = b[x+8];
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    w3 = *(unsigned short *) (&b[x+10]);
                    if (c1 & 4)
                        fprintf(outFile,"anim id in Param %d", w1);
                    else
                        fprintf(outFile,"anim id %X", w1);
                    fprintf(outFile," from ");
                    switch (b[x+6]) {
                        case 0x80:
                            fprintf(outFile,"Global list");
                            break;
                        case 0x81:
                            fprintf(outFile,"Adult list");
                            break;
                        case 0x82:
                            fprintf(outFile,"Child list");
                            break;
                        case 0x83:
                            fprintf(outFile,"Social Interaction list");
                            break;
                        case 0x89:
                            fprintf(outFile,"Toddler list");
                            break;
                        case 0x8A:
                            fprintf(outFile,"Teen list");
                            break;
                        case 0x8B:
                            fprintf(outFile,"Elder list");
                            break;
                        case 0x8C:
                            fprintf(outFile,"Cat list");
                            break;
                        case 0x8D:
                            fprintf(outFile,"Dog list");
                            break;
                        case 0x91:
                            fprintf(outFile,"Baby list");
                            break;
                        default:
                            fprintf(outFile,"Object list");
                            break;
                    }
                    if (~(c1 & 4)) {
                        fprintf(outFile," (");
                        if (b[x+6] == 0x80) {
                            if (readString2(0x7FD46CD0, 0x81, w1) == 0)
                                fprintf(outFile,"Global STR# 0x81:%X", 0x81, w1);
                        } else if ((b[x+6] > 0x80 && b[x+6] < 0x84) || (b[x+6] > 0x88 && b[x+6] < 0x8E) || b[x+6] == 91) {
                            if (readString2(gGroup, b[x+6], w1) == 0)
                                  fprintf(outFile,"STR# 0x%X:%X", b[x+6], w1);
                        } else
                            if (readString2(gGroup, 0x86, w1) == 0)
                                fprintf(outFile,"STR# 0x86:%X", 0x86, w1);
                        fprintf(outFile,")");
                    }
                    if (w2 == 0)
                        fprintf(outFile,", No Event Tree");
                    else {
                        switch (b[x+7]) {
                            case 0:
                                fprintf(outFile,", Private Event Tree: ");
                                break;
                            case 1:
                                fprintf(outFile,", SemiGlobal Event Tree: ");
                                break;
                            default:
                                fprintf(outFile,", Global Event Tree: ");
                                break;
                        }
                        fprintf(outFile,"0x%X",w2);
                        fprintf(outFile," (");
                        readFn2(w2);
                        fprintf(outFile,")");
                    }
                    if (c1 & 8)
                        fprintf(outFile,", Interruptible");
                    if (c1 & 1)
                        fprintf(outFile,", Flipped");
                    if (c1 & 0x20)
                        fprintf(outFile,", Trans to Idle");
                    if (c1 & 0x80)
                        fprintf(outFile,", No Blend in");
                    if (c1 & 0x40)
                        fprintf(outFile,", No Blend out");
                    if (c1 & 0x10)
                        fprintf(outFile,", Start at tag in Temp 0");
                    if (c1 & 2)
                        fprintf(outFile,", Anim Speed in Temp2");
                    if (c2 & 1)
                        fprintf(outFile,", Flip Flag in Temp 3");
                    if (c2 & 2)
                        fprintf(outFile,", Synching Anim to last Anim Run");
                    if (c2 & 0x10)
                        fprintf(outFile,", Using Controlling Object as Anim Source");
                    if (c2 & 0x20)
                        fprintf(outFile,", Not Hurryable");
                    fprintf(outFile,", IK Object ");
                    data2(b[x+9],w3);
                    switch (b[x+12]) {
                        case 0:
                            fprintf(outFile,", low priority");
                            break;
                        case 1:
                            fprintf(outFile,", medium priority");
                            break;
                        case 2:
                            fprintf(outFile,", high priority");
                            break;
                    }
                    break;
                case 0x6B:  // Animate Overlay (false = error)
                    c1 = b[x+2];
                    c2 = b[x+15];
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    w3 = *(unsigned short *) (&b[x+7]);
                    if (c1 & 4)
                        fprintf(outFile,"anim id in Param %d", w1);
                    else
                        fprintf(outFile,"anim id %X", w1);
                    fprintf(outFile," from ");
                    switch (b[x+6]) {
                        case 0x80:
                            fprintf(outFile,"Global list");
                            break;
                        case 0x81:
                            fprintf(outFile,"Adult list");
                            break;
                        case 0x82:
                            fprintf(outFile,"Child list");
                            break;
                        case 0x83:
                            fprintf(outFile,"Social Interaction list");
                            break;
                        case 0x89:
                            fprintf(outFile,"Toddler list");
                            break;
                        case 0x8A:
                            fprintf(outFile,"Teen list");
                            break;
                        case 0x8B:
                            fprintf(outFile,"Elder list");
                            break;
                        case 0x8C:
                            fprintf(outFile,"Cat list");
                            break;
                        case 0x8D:
                            fprintf(outFile,"Dog list");
                            break;
                        case 0x91:
                            fprintf(outFile,"Baby list");
                            break;
                        default:
                            fprintf(outFile,"Object list");
                            break;
                    }
                    if (~(c1 & 4)) {
                        fprintf(outFile," (");
                        if (b[x+9] == 0x80) {
                            if (readString2(0x7FD46CD0, 0x81, w1) == 0)
                                fprintf(outFile,"Global STR# 0x81:%X", 0x81, w1);
                        } else if ((b[x+9] > 0x80 && b[x+9] < 0x84) || (b[x+9] > 0x88 && b[x+9] < 0x8E) || b[x+9] == 91) {
                            if (readString2(gGroup, b[x+9], w1) == 0)
                                  fprintf(outFile,"STR# 0x%X:%X", b[x+9], w1);
                        } else
                            if (readString2(gGroup, 0x86, w1) == 0)
                                fprintf(outFile,"STR# 0x86:%X", 0x86, w1);
                        fprintf(outFile,")");
                    }
                    fprintf(outFile," affecting ");
                    data2(b[x+6],w3);       // target object
                    if (w2 == 0)
                        fprintf(outFile,", No Event Tree");
                    else {
                        switch (b[x+14]) {
                            case 0:
                                fprintf(outFile,", Private Event Tree: ");
                                break;
                            case 1:
                                fprintf(outFile,", SemiGlobal Event Tree: ");
                                break;
                            default:
                                fprintf(outFile,", Global Event Tree: ");
                                break;
                        }
                        fprintf(outFile,"0x%X",w2);
                        fprintf(outFile," (");
                        readFn2(w2);
                        fprintf(outFile,")");

                    }
                    if (c1 & 8)
                        fprintf(outFile,", Interruptible");
                    if (c1 & 1)
                        fprintf(outFile,", Flipped");
                    if (c1 & 0x80)
                        fprintf(outFile,", No Blend in");
                    if (c1 & 0x40)
                        fprintf(outFile,", No Blend out");
                    if (c1 & 0x10)
                        fprintf(outFile,", Start at tag in Temp 0");
                    if (c1 & 0x20)
                        fprintf(outFile,", Loop Count in Temp 1");
                    if (c1 & 2)
                        fprintf(outFile,", Anim Speed in Temp2");
                    if (c2 & 1)
                        fprintf(outFile,", Flip Flag in Temp 3");
                    if (c2 & 0x10)
                        fprintf(outFile,", Sync to calling objects Anim");
                    if (c2 & 0x20)
                        fprintf(outFile,", Align blend out with to calling objects Anim");
                    if (c2 & 0x40)
                        fprintf(outFile,", Using tree owners anim list");
                    if (b[x+12] & 1)
                        fprintf(outFile,", Not Hurryable");
                    switch (b[x+11]) {
                        case 0:
                            fprintf(outFile,", low priority");
                            break;
                        case 1:
                            fprintf(outFile,", medium priority");
                            break;
                        case 2:
                            fprintf(outFile,", high priority");
                            break;
                    }
                    break;
                case 0x6C:  // Animate Stop (false = error)
                    c1 = b[x+2];
                    w1 = *(unsigned short *) (&b[x]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    switch (b[x+7]) {
                        case 0:
                            if (c1 & 4)
                                fprintf(outFile,"anim id in Param %d", w1);
                            else
                                fprintf(outFile,"anim id %X", w1);
                            fprintf(outFile," from ");
                            switch (b[x+6]) {
                                case 0x80:
                                    fprintf(outFile,"Global list");
                                    break;
                                case 0x81:
                                    fprintf(outFile,"Adult list");
                                    break;
                                case 0x82:
                                    fprintf(outFile,"Child list");
                                    break;
                                case 0x83:
                                    fprintf(outFile,"Social Interaction list");
                                    break;
                                case 0x89:
                                    fprintf(outFile,"Toddler list");
                                    break;
                                case 0x8A:
                                    fprintf(outFile,"Teen list");
                                    break;
                                case 0x8B:
                                    fprintf(outFile,"Elder list");
                                    break;
                                case 0x8C:
                                    fprintf(outFile,"Cat list");
                                    break;
                                case 0x8D:
                                    fprintf(outFile,"Dog list");
                                    break;
                                case 0x91:
                                    fprintf(outFile,"Baby list");
                                    break;
                                default:
                                    fprintf(outFile,"Object list");
                                    break;
                            }
                            if (~(c1 & 4)) {
                                fprintf(outFile," (");
                                if (b[x+6] == 0x80) {
                                    if (readString2(0x7FD46CD0, 0x81, w1) == 0)
                                        fprintf(outFile,"Global STR# 0x81:%X", 0x81, w1);
                                } else if ((b[x+6] > 0x80 && b[x+6] < 0x84) || (b[x+6] > 0x88 && b[x+6] < 0x8E) || b[x+6] == 91) {
                                    if (readString2(gGroup, b[x+6], w1) == 0)
                                          fprintf(outFile,"STR# 0x%X:%X", b[x+6], w1);
                                } else
                                    if (readString2(gGroup, 0x86, w1) == 0)
                                        fprintf(outFile,"STR# 0x86:%X", 0x86, w1);
                                fprintf(outFile,")");
                            }
                            break;
                        case 1:
                            fprintf(outFile,"All Overlay animations running");
                            break;
                        case 2:
                            fprintf(outFile,"All Full Body Animations running");
                            break;
                        case 3:
                            fprintf(outFile,"All Animations running");
                            break;
                        default:
                            fprintf(outFile,"Carry Pose");
                    }
                    fprintf(outFile," affecting ");
                    data2(b[x+3],w2);       // target object
                    if (c1 & 1)
                        fprintf(outFile,", Flipped");
                    if (c1 & 2)
                        fprintf(outFile,", blended Out");
                    if (c1 & 0x20)
                        fprintf(outFile,", short blended Out");
                    if (c1 & 8)
                        fprintf(outFile,", Flip Flag in Temp 3");
                    if (c1 & 0x40)
                        fprintf(outFile,", Stopping normal AND flipped anims");
                    fprintf(outFile,", Priority ");
                    switch (b[x+8]) {
                        case 0:
                            fprintf(outFile,"Any");
                            break;
                        case 1:
                            fprintf(outFile,"Low");
                            break;
                        case 2:
                            fprintf(outFile,"Medium");
                            break;
                        default:
                            fprintf(outFile,"High");
                    }
                    break;
                case 0x6D:  // Change Material (false = error)
                    w1 = *(unsigned short *) (&b[x+6]);
                    w2 = *(unsigned short *) (&b[x+9]);
                    w3 = *(unsigned short *) (&b[x+3]);
                    w4 = *(unsigned short *) (&b[x]);
                    c1 = b[x+2];
                    if (b[x+13] & 2) {
                        fprintf(outFile,"using snap shot generated material");
                        if (c1 & 1) {
                            fprintf(outFile,", mesh found in obj in");
                            data2(b[x+8],w2);
                            if (c1 & 0x20)
                                fprintf(outFile,", using mesh group index from temp 1");
                            else
                                fprintf(outFile,", using mesh group index %d", w3);
                        } else if (c1 & 0x20)
                            fprintf(outFile,", using mesh group index from temp 1");
                        else if (b[x+4] & 0x40) // w3 < 0
                            fprintf(outFile,", over all model");
                        else {
                            fprintf(outFile,", On Mesh Group: ");
                            if (c1 & 0x40)
                                readString2(0x7FD46CD0, 0x87, w3);
                            else if (c1 & 0x80) {
                                if (readString2(gGlobGroup, 0x87, w3) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x87:0x%X]", w3);
                            }
                            else if (readString2(gGroup, 0x87, w3) == 0)
                                fprintf(outFile,"[Private STR# 0x87:0x%X]", w3);
                        }
                    } else if (c1 & 8) {
                        fprintf(outFile,"found in obj in ");
                        data2(b[x+8],w2);
                        if (c1 & 0x10)
                            if (b[x+13] & 1)
                                fprintf(outFile,", using Moving Texture Name index from temp 0");
                            else
                                fprintf(outFile,", using material index from temp 0");
                        else
                            if (b[x+13] & 1)
                                fprintf(outFile,", using Moving Texture Name index %d", w4);
                            else
                                fprintf(outFile,", using material index %d", w4);
                    } else {
                        fprintf(outFile,"found in me, to: ");
                        if (c1 & 0x10)
                            if (b[x+13] & 1)
                                fprintf(outFile,", using Moving Texture Name index from temp 0");
                            else
                                fprintf(outFile,", using material index from temp 0");
                        else
                            if (c1 & 2)
                                readString2(0x7FD46CD0, 0x88, w4);
                            else if (c1 & 4) {
                                if (readString2(gGlobGroup, 0x88, w4) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x88:0x%X]", w4);
                            }
                            else if (readString2(gGroup, 0x88, w4) == 0)
                                fprintf(outFile,"[Private STR# 0x88:0x%X]", w4);
                    }
                    fprintf(outFile,", affecting ID in ");
                    data2(b[x+5],w1);       // target object
                    break;
                case 0x6E:  // Look At
                    w1 = *(unsigned short *) (&b[x+2]);
                    w2 = *(unsigned short *) (&b[x+9]);
                    c1 = b[x];
                    if (~(b[x] & 1)) {
                        if (w2 == 0)
                            fprintf(outFile,"No Event Tree");
                        else {
                            switch (b[x+11]) {
                                case 0:
                                    fprintf(outFile,"Private Event Tree: ");
                                    break;
                                case 1:
                                    fprintf(outFile,"SemiGlobal Event Tree: ");
                                    break;
                                default:
                                    fprintf(outFile,"Global Event Tree: ");
                                    break;
                            }
                            fprintf(outFile,"0x%X",w2);
                            fprintf(outFile," (");
                            readFn2(w2);
                            fprintf(outFile,")");
                        }
                        if (~(b[x] & 0x80)) {
                            fprintf(outFile,", Targeting ID in ");
                            data2(b[x+1],w1);
                        } else
                            fprintf(outFile,", Targeting Camera ");
                        fprintf(outFile," using ");
                        switch (b[x+14]) {
                            case 0:
                                fprintf(outFile,"default 3/4 height");
                                break;
                            case 1:
                                fprintf(outFile,"targeting slot"," number %d",b[x+8]);
                                break;
                            case 2:
                                fprintf(outFile,"routing slot"," number %d",b[x+8]);
                                break;
                            default:
                                fprintf(outFile," containment slot"," number %d",b[x+8]);
                                break;
                        }
                        if (c1 & 4)
                            fprintf(outFile,", including Spine");
                        if (c1 & 0x10)
                            fprintf(outFile,", using duration in temp 0");
                        if (c1 & 2)
                            fprintf(outFile,", no early exit");
                    } else
                        fprintf(outFile,"STOP");
                    if (c1 & 8)
                        fprintf(outFile,", using turnTowardsSpeed in temp 1");
                    else
                        fprintf(outFile,", turn towards speed of %d  degrees per second",2*b[x+4]);
                    if (b[x+15] & 2)
                        fprintf(outFile,", using turnAwaySpeed in temp 1");
                    else if (b[x+15] & 1)
                        fprintf(outFile,", using turnAwaySpeed in temp 2");
                    else
                        fprintf(outFile,", turn away speed of %d  degrees per second",2*b[x+5]);
                    if (c1 & 0x20)
                        fprintf(outFile,", ignoring Room");
                    if (c1 & 0x40)
                        fprintf(outFile,", ignoring Frustrum");
                    if (b[x+15] & 4)
                        fprintf(outFile,", Not Hurryable");
                    break;
                case 0x6F:  // Change Light (false = error)
                    w1 = *(unsigned short *) (&b[x+3]);
                    w2 = *(unsigned short *) (&b[x+5]);
                    fprintf(outFile,"on object in ");
                    data2(b[x+2],w1);       // target object
                    if (b[x+8] != 0xFF) {
                        fprintf(outFile,", Targeting light ");
                        if (readString2(gGroup, 0x8E, b[x+8]) == 0)
                            fprintf(outFile,"[STR# 0x8E:0x%X]", b[x+8]);
                    } else
                        fprintf(outFile,", Targeting all lights on Object");
                    if (b[x+1] & 1)
                        fprintf(outFile,", Fade duration in Temp 1,");
                    else
                        fprintf(outFile,", Fading over a duration of %d ticks,", w2);
                    if (b[x+1] & 2)
                        fprintf(outFile," Intensity in Temp 0");
                    else
                        fprintf(outFile," Intensity of %d percent", b[x+7]);
                    break;
                case 0x70:  // Effect Stop/Start (false = error)
                    w1 = *(unsigned short *) (&b[x+2]);
                    w2 = *(unsigned short *) (&b[x+13]);
                    switch (b[x]) {
                        case 0:
                            fprintf(outFile,"Soft Start Effect on object in ");
                            break;
                        case 1:
                            fprintf(outFile,"Hard start effect on object in ");
                            break;
                        case 2:
                            fprintf(outFile,"Soft stop effect on object in ");
                            break;
                        case 3:
                            fprintf(outFile,"Hard stop effect on object in ");
                            break;
                        case 4:
                            fprintf(outFile,"Soft stop all effects on object in ");
                            break;
                        case 5:
                            fprintf(outFile,"Hard stop all effects on object in ");
                            break;
                        case 6:
                            fprintf(outFile,"Fire and Forget Effect on object in ");
                            break;
                        case 7:
                            fprintf(outFile,"Interrogate Bone for effects on object in ");
                            break;
                        case 8:
                            fprintf(outFile,"Hard stop all effects and clear Queue on object in ");
                            break;
                        case 9:
                            fprintf(outFile,"Hard stop ALL effects on object in ");
                            break;
                        case 0xA:
                            fprintf(outFile,"Set State 1 for all effects on object in ");
                            break;
                        case 0xB:
                            fprintf(outFile,"Set State 2 for all effects on object in ");
                            break;
                        case 0xC:
                            fprintf(outFile,"Set State 3 for all effects on object in ");
                            break;
                        case 0xD:
                            fprintf(outFile,"Set State 4 for all effects on object in ");
                            break;
                        case 0xE:
                            fprintf(outFile,"Soft stop ALL effects on object in ");
                            break;
                    }
                    data2(b[x+1],w1);       // target object
                    if (b[x] == 4 || b[x] == 5) {
                        if (b[x+10] & 0x40)
                            fprintf(outFile,", Passing in effect ID in temp 1");
                    } else if (b[x] < 7 || b[x] == 0xE)
                        if (b[x+4] != 0xFF) {
                            fprintf(outFile,", ");
                            if (b[x+10] & 1)
                                readString2(0x7FD46CD0, 0x8F, b[x+4]); // !!!
                            else if (b[x+10] & 2) {
                                if (readString2(gGlobGroup, 0x8F, b[x+4]) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x8F:0x%X]", b[x+4]);
                            }
                            else if (readString2(gGroup, 0x8F, b[x+4]) == 0)
                                fprintf(outFile,"[Private STR# 0x8F:0x%X]", b[x+4]);
                        } else
                            fprintf(outFile,", Affecting default effect");
                    if (b[x] != 9) {
                        fprintf(outFile,", of Slot Type ");
                        switch (b[x+9]) {
                            case 0:
                                fprintf(outFile,"Target");
                                break;
                            case 1:
                                fprintf(outFile,"Routing");
                                break;
                            default:
                                fprintf(outFile,"Containment");
                                break;
                        }
                        fprintf(outFile,", using slot number %d", b[x+6]);
                    }
                    if (b[x+11] & 0x10)
                        fprintf(outFile,", getting icon value from Temp6");
                    else if (b[x+11] & 4) {
                        fprintf(outFile,", putting in Icon with GUID in temp4/5 ");
                        if (b[x+11] & 8)
                            fprintf(outFile," getting model name index from Temp6");
                        else
                            fprintf(outFile," using default object model");
                    } else if (b[x+10] & 0x10) {
                        fprintf(outFile,", putting in Icon from neighbor ID in ");
                        data2(b[x+12],w2);
                    } else if (b[x+10] & 0x20) {
                        fprintf(outFile,", putting in Conversation Icon index found in ");
                        data2(b[x+12],w2);
                        fprintf(outFile," using sheet ");
                        if (readString2(gGroup, 0x95, b[x+15]) == 0) // ??
                            fprintf(outFile,"[STR# 0x95:0x%X]", b[x+15]);
                    } else if (b[x+10] & 4) {
                        fprintf(outFile,", putting in Icon from object in ");
                        data2(b[x+12],w2);
                        if (b[x+11] & 8)
                            fprintf(outFile," getting model name index from Temp6");
                        else
                            fprintf(outFile," using default object model");
                    }
                    if (b[x+10] & 0x80)
                        fprintf(outFile,", putting effect in priority Queue");
                    break;
                case 0x71:  // Snap Into
                    w1 = *(unsigned short *) (&b[x+1]);
                    w2 = *(unsigned short *) (&b[x+4]);
                    fprintf(outFile,"Snap Object in ");
                    data2(b[x],w1);
                    fprintf(outFile," into Object in ");
                    data2(b[x+3],w2);
                    if (b[x+9] & 1)
                        fprintf(outFile,", using slot # in temp 0");
                    else
                        fprintf(outFile,", using slot number %d", b[x+6]);
                    if (b[x+9] & 2)
                        fprintf(outFile,", TEST ONLY");
                    if (b[x+9] & 4)
                        fprintf(outFile,", Reset root bones");
                    break;
                case 0x72:  // Assign Locomotion Animations (false = error)
                    w1 = *(unsigned short *) (&b[x]);
                    switch (w1) {
                        case 0:
                            fprintf(outFile,"Pop all entries from Stack.");
                            break;
                        case 1:
                            fprintf(outFile,"Pop last entry from Stack.");
                            break;
                        default:
                            fprintf(outFile,"Push group #%d onto Stack", w1-2);
                            fprintf(outFile,", getting group from ");
                            if (b[x+2] & 4)
                                fprintf(outFile,"Global.");
                            else if (b[x+2] & 2)
                                fprintf(outFile,"SemiGlobal.");
                            else
                                fprintf(outFile,"Private.");
                    }
                    break;
                case 0x73:  // Debug (false = error)
                    switch (b[x+13]) {
                        case 0:
                            if (b[x+14] & 4)
                                readString2(0x7FD46CD0, 0x132, b[x+12]);
                            else if (b[x+14] & 2) {
                                if (readString2(gGlobGroup, 0x132, b[x+12]) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x132:0x%X]", b[x+12]);
                            }
                            else if (readString2(gGroup, 0x132, b[x+12]) == 0)
                                fprintf(outFile,"[Private STR# 0x132:0x%X]", b[x+12]);
                            w1 = *(unsigned short *) (&b[x+1]);
                            w2 = *(unsigned short *) (&b[x+4]);
                            w3 = *(unsigned short *) (&b[x+7]);
                            w4 = *(unsigned short *) (&b[x+10]);
                            fprintf(outFile,", param 0: ");
                            data2(b[x], w1);
                            fprintf(outFile,", param 1: ");
                            data2(b[x+3], w2);
                            fprintf(outFile,", param 2: ");
                            data2(b[x+6], w3);
                            fprintf(outFile,", param 3: ");
                            data2(b[x+9], w4);
                            break;
                        case 1:
                            fprintf(outFile,"Toggle Window Open/Close");
                            break;
                        case 2:
                            fprintf(outFile,"Open Animation Ticker");
                            break;
                        case 3:
                            fprintf(outFile,"Show Slots");
                            break;
                        case 4:
                            fprintf(outFile,"Show Bones");
                            break;
                        case 5:
                            fprintf(outFile,"Toggle Anim Info to Debug Window");
                            break;
                        case 6:
                            fprintf(outFile,"Perform Cheat ");
                            if (b[x+14] & 4)
                                readString2(0x7FD46CD0, 0x132, b[x+12]);
                            else if (b[x+14] & 2) {
                                if (readString2(gGlobGroup, 0x132, b[x+12]) == 0)
                                    fprintf(outFile,"[SemiGlobal STR# 0x132:0x%X]", b[x+12]);
                            }
                            else if (readString2(gGroup, 0x132, b[x+12]) == 0)
                                fprintf(outFile,"[Private STR# 0x132:0x%X]", b[x+12]);
                            break;
                        case 7:
                            fprintf(outFile,"Dump Happy Log");
                            break;
                    }
                    break;
                case 0x74:  // Reach/Put
                    w1 = *(unsigned short *) (&b[x+4]);
                    w2 = *(unsigned short *) (&b[x+13]);
                    w3 = *(unsigned short *) (&b[x+11]);
                    switch (b[x+10]) {
                        case 0:
                            fprintf(outFile,"Pick up Object in ");
                            data2(b[x+3],w1);
                            break;
                        case 1:
                            fprintf(outFile,"Drop object onto Floor");
                            break;
                        default:
                            fprintf(outFile,"Drop onto Object in ");
                            data2(b[x+3],w1);
                    }
                    if (b[x+10]!= 1 && b[x+10]!= 2)
                        if (b[x+9] & 1)
                            fprintf(outFile,", using slot # in temp 0");
                        else
                            fprintf(outFile,", using slot number %d", b[x+6]);
                    fprintf(outFile,", object anim: ");
                    if (w2 != 0xFFFF) {
                        if (readString2(gGroup, 0x86, w2) == 0) // ?
                            fprintf(outFile,"[STR# 0x86:0x%X]", w2);
                    } else
                        fprintf(outFile,"none");
                    if (b[x+9] & 4)
                        fprintf(outFile,", using Sim age as filter on object Anim");
                    fprintf(outFile,", grasp anim: ");
                    if (w3 != 0xFFFF) {
                        if (readString2(gGroup, 0x81, w3) == 0) // ?
                            fprintf(outFile,"[STR# 0x81:0x%X]", w3);
                    } else
                        fprintf(outFile,"none");
                    if (b[x+9] & 2)
                        fprintf(outFile,", expecting handedness in temp 3");
                    break;
                case 0x75:  // Age (false = error)
                    if (b[x+1] & 1)
                        fprintf(outFile,"Expect Age in Temp 0");
                    else {
                        fprintf(outFile,"Set Age to ");
                        switch (b[x]) {
                            case 1:
                                fprintf(outFile,"Child");
                                break;
                            case 2:
                                fprintf(outFile,"Toddler");
                                break;
                            case 3:
                                fprintf(outFile,"Teen");
                                break;
                            case 4:
                                fprintf(outFile,"Elder");
                                break;
                            case 7:
                                fprintf(outFile,"Baby");
                                break;
                            default:
                                fprintf(outFile,"Adult");
                        }
                    }
                    break;
                case 0x76:  // Array Operation
                    w1 = *(unsigned short *) (&b[x+6]);
                    w2 = *(unsigned short *) (&b[x+9]);
                    w3 = *(unsigned short *) (&b[x+3]);
                    if (b[x+2] == 0)
                        fprintf(outFile,"My Object Array: ");
                    else
                        fprintf(outFile,"Stack Object's Object Array: ");
                    if (readString2(gGroup, 0x118, w3) == 0)
                                fprintf(outFile,"[STR# 0x118:0x%X]", w3);
                    fprintf(outFile,". Operation: ");
                    switch (b[x+1]) {
                        case 0:
                            fprintf(outFile,"Clear contents of array.");
                            break;
                        case 1:
                            fprintf(outFile,"Return size of array into ");
                            data2(b[x+5],w1);
                            fprintf(outFile,".");
                            break;
                        case 2:
                            fprintf(outFile,"Resize array to size from ");
                            data2(b[x+5],w1);
                            fprintf(outFile,".");
                            break;
                        case 3:
                            fprintf(outFile,"Init all elements of the array with value from ");
                            data2(b[x+5],w1);
                            fprintf(outFile,".");
                            break;
                        case 4:
                            fprintf(outFile,"Insert element to array from ");
                            data2(b[x+5],w1);
                            fprintf(outFile," to the front of the array.");
                            break;
                        case 5:
                            fprintf(outFile,"Insert element to array from ");
                            data2(b[x+5],w1);
                            fprintf(outFile," to the back of the array.");
                            break;
                        case 6:
                            fprintf(outFile,"Insert element to array from ");
                            data2(b[x+5],w1);
                            fprintf(outFile," to the position indicated by iterator in ");
                            data2(b[x+8],w2);
                            fprintf(outFile,".");
                            break;
                        case 7:
                            fprintf(outFile,"Remove front element from array.");
                            break;
                        case 8:
                            fprintf(outFile,"Remove back element from array.");
                            break;
                        case 9:
                            fprintf(outFile,"Remove element from array at position given by iterator in ");
                            data2(b[x+8],w2);
                            fprintf(outFile,".");
                            break;
                        case 0xA:
                            fprintf(outFile,"Advance iterator in ");
                            data2(b[x+8],w2);
                            fprintf(outFile," to next occurance of element from ");
                            data2(b[x+5],w1);
                            fprintf(outFile,".");
                            break;
                        case 0xB:
                            fprintf(outFile,"Swap elements of array at iterator ");
                            data2(b[x+5],w1);
                            fprintf(outFile," and iterator ");
                            data2(b[x+8],w2);
                            fprintf(outFile,".");
                            break;
                        case 0xC:
                            fprintf(outFile,"Sort array into highest to lowest order.");
                            break;
                        case 0xD:
                            fprintf(outFile,"Sort array into highest to lowest order.");
                            break;
                        default:
                            fprintf(outFile,"Alex Fennell's magic array operation. If you see this string....be afraid.");
                    }
                    break;
                case 0x77:  // Message
                    w1 = *(unsigned short *) (&b[x+1]);
                    w2 = *(unsigned short *) (&b[x+6]);
                    w3 = *(unsigned short *) (&b[x+10]);
                    w4 = *(unsigned short *) (&b[x+13]);
                    fprintf(outFile,"Sending Message ID inside ");
                    data2(b[x+15],w1);
                    fprintf(outFile,", priority %d",b[x+8]);
                    if (b[x+4] & 4) {
                        fprintf(outFile," Targetting specific target in ");
                        data2(b[x+5],w2);
                    } else {
                        fprintf(outFile," Targetting ");
                        switch (b[x+3]) {
                            case 0:
                                fprintf(outFile,"Selectable Sims");
                                break;
                            case 1:
                                fprintf(outFile,"Selectable Sims + Neighbors");
                                break;
                            case 2:
                                fprintf(outFile,"Selectable Sims + NPCS");
                                break;
                            case 3:
                                fprintf(outFile,"Neighbors Only");
                                break;
                            case 4:
                                fprintf(outFile,"NPCS Only");
                                break;
                            case 5:
                                fprintf(outFile,"All Sims");
                                break;
                            case 6:
                                fprintf(outFile,"Objects");
                                break;
                            case 7:
                                fprintf(outFile,"Everything");
                                break;
                        }
                        fprintf(outFile," using filter ");
                        switch (b[x]) {
                            case 0:
                                fprintf(outFile,"In Same Room Only");
                                if (b[x+4] & 1) {
                                    fprintf(outFile," targeting room in ");
                                    data2(b[x+5],w2);
                                }
                                break;
                            case 1:
                                fprintf(outFile,"On Same Level Only");
                                break;
                            case 2:
                                fprintf(outFile,"On Lot");
                                break;
                            case 3:
                                fprintf(outFile,"Inside building Only");
                                break;
                            case 4:
                                fprintf(outFile,"OutSide Building Only");
                                break;
                        }
                        fprintf(outFile,", getting user data from ");
                        data2(b[x+9],w3);
                        fprintf(outFile,", and ");
                        data2(b[x+12],w4);
                    }
                    break;
                case 0x78:  // RayTrace
                    w1 = *(unsigned short *) (&b[x+2]);
                    w2 = *(unsigned short *) (&b[x+9]);
                    fprintf(outFile,"From Object in ");
                    data2(b[x+1],w1);
                    fprintf(outFile," using ");
                    switch (b[x+4]) {
                        case 0:
                            fprintf(outFile,"point at 3/4 height of object.");
                            break;
                        case 1:
                            fprintf(outFile,"targeting slot");
                            break;
                        case 2:
                            fprintf(outFile,"routing slot");
                            break;
                        default:
                            fprintf(outFile,"containment slot");
                    }
                    if (b[x+4] != 0)
                        fprintf(outFile,", slot number %d into the world. Object hit stored in Temp 0.", b[x+5]);
                    else {
                        fprintf(outFile," To Object in ");
                        data2(b[x+8],w2);
                        fprintf(outFile," using ");
                        switch (b[x+11]) {
                            case 0:
                                fprintf(outFile,"point at 3/4 height of object.");
                                break;
                            case 1:
                                fprintf(outFile,"targeting slot");
                                break;
                            case 2:
                                fprintf(outFile,"routing slot");
                                break;
                            default:
                                fprintf(outFile,"containment slot");
                        }
                        if (b[x+11] != 0)
                            fprintf(outFile,", slot number %d.", b[x+12]);
                    }
                    if (b[x+15] & 1)
                        fprintf(outFile," Windows Ignored.");
                    break;
                case 0x79:  // Change Outfit (false = error)
                    c1 = b[x];
                    w1 = *(unsigned short *) (&b[x+10]);
                    w2 = *(unsigned short *) (&b[x+2]);
                    d1 = *(unsigned long *) (&b[x+4]);
                    if (c1 & 0x10) {
                        fprintf(outFile,"Rebuild current outfit on sim in ");
                        data2(b[x+9],w1);
                    } else {
                        fprintf(outFile,"Change Outfit on sim in ");
                        data2(b[x+9],w1);
                        if (c1 & 1)
                            fprintf(outFile," using Stack Object");
                        else if (c1 & 2) {
                            fprintf(outFile," using GUID of 0x%08X", d1);
                            readGUID(d1);
                        } else if (c1 & 0x40)
                            fprintf(outFile,", using GUID in Temp 0/1");
                        else
                            fprintf(outFile," using that sims outfits");
                        fprintf(outFile," as source, using outfit ");
                        if (c1 & 4) {
                            fprintf(outFile,"index from ");
                            data2(b[x+1],w2);
                        } else
                            fprintf(outFile,"%s", gPersonOutfits[b[x+8]]);
                        if (c1 & 0x20)
                            fprintf(outFile,", clearing GUID pointers in person data fields");
                        if (c1 & 8)
                            fprintf(outFile,", writing changes to the .iff");
                    }
                    break;
                case 0x7A:  // On Timer (false = error)
                    w1 = *(unsigned short *) (&b[x+7]);
                    w2 = *(unsigned short *) (&b[x+10]);
                    w3 = *(unsigned short *) (&b[x+13]);
                    w4 = *(unsigned short *) (&b[x]);
                    w5 = *(unsigned short *) (&b[x+3]);
                    switch (b[x+15]) {
                        case 0:
                            fprintf(outFile,"Start Timer");
                            break;
                        case 1:
                            fprintf(outFile,"Modify Timer");
                            break;
                        case 2:
                            fprintf(outFile,"Delete Timer");
                            break;
                    }
                    if (b[x+15] != 2) {
                        if (b[x+5] & 8)
                            fprintf(outFile,", Getting ticks till fire from temp 1");
                        else
                            fprintf(outFile,", Ticks till fire %d", w4);

                        if (w5 == 0)
                            fprintf(outFile," No Event Tree");
                        else {
                            switch (b[x+2]) {
                                case 0:
                                    fprintf(outFile," Using Private Event Tree: ");
                                    break;
                                case 1:
                                    fprintf(outFile," Using SemiGlobal Event Tree: ");
                                    break;
                                default:
                                    fprintf(outFile," Using Global Event Tree: ");
                                    break;
                            }
                            fprintf(outFile,"0x%X",w5);
                            fprintf(outFile," (");
                            readFn2(w5);
                            fprintf(outFile,")");
                         }
                        if (b[x+5] & 2)
                            fprintf(outFile,", Looping");
                        if (b[x+5] & 1)
                            fprintf(outFile,", using current params as event tree params");
                        else {
                            fprintf(outFile,", passing in params found in  ");
                            data2(b[x+6],w1);
                            fprintf(outFile,", ");
                            data2(b[x+9],w2);
                            fprintf(outFile,", ");
                            data2(b[x+12],w3);
                        }
                        if (b[x+15] == 1)
                            if (b[x+5] & 4)
                                fprintf(outFile,", Reset Timer Ticks");
                    }
                    break;
                case 0x7B:  // Cinematic
                    c1 = b[x+5];
                    w1 = *(unsigned short *) (&b[x+7]);
                    w2 = *(unsigned short *) (&b[x]);
                    w3 = *(unsigned short *) (&b[x+10]);
                    if (c1 & 0x10) {
                        fprintf(outFile,"Using scene name index found in ");
                        data2(b[x+6],w1);
                    } else {
                        fprintf(outFile,"Using Scene ");
                        if (c1 & 0x20)
                            readString2(0x7FD46CD0, 0x97, w2);
                        else if (c1 & 0x40) {
                            if (readString2(gGlobGroup, 0x97, w2) == 0)
                                fprintf(outFile,"[SemiGlobal STR# 0x97:0x%X]", w2);
                        }
                        else if (readString2(gGroup, 0x97, w2) == 0)
                            fprintf(outFile,"[Private STR# 0x97:0x%X]", w2);
                    }
                    fprintf(outFile,", Using array found in objID in ");
                    data2(b[x+9],w3);
                    if (c1 & 1)
                        fprintf(outFile,", Flip Cinematic Anims Horizontally");
                    if (c1 & 2)
                        fprintf(outFile,", Pass in Flip Cinematic Anims Horizontally value in temp 0");
                    if (c1 & 4)
                        fprintf(outFile,", Start Animations Now");
                    if (c1 & 8)
                        fprintf(outFile,", Show Entire House");
                    break;
                case 0x7C:  // Want Satisfy
                    w3 = *(unsigned short *) (&b[x+8]);
                    w4 = *(unsigned short *) (&b[x+11]);
                    w5 = *(unsigned short *) (&b[x+1]);
                    d1 = *(unsigned long *) (&b[x+3]);
                    fprintf(outFile,"GUID 0x%08X ", d1);
                    w2 = 0xFFFF;
                    for (w1 = 0; w1 < NUM_WANTS; w1++) if (d1 == gWantsIDs[w1]) w2 = w1;
                    if (w2 != 0xFFFF) {
                        fprintf(outFile,"(%s)", gWantsNames[w2]);
                        switch (gWantsTypes[w2]) {
                            case 0:        // None
                                fprintf(outFile," with value ");
                                break;
                            case 1:        // Sim
                                fprintf(outFile," with person in ");
                                break;
                            case 2:        // Guid
                            case 3:        // Category
                                fprintf(outFile," with object in ");
                                break;
                            case 4:        // Skill
                                fprintf(outFile," with skill ");
                                break;
                            case 5:        // Career
                                fprintf(outFile," with career ");
                                break;
                        }
                        data2(b[x+7],w3);
                        if (gWantsTypes[w2] == 4 || gWantsTypes[w2] == 5) {
                            fprintf(outFile," and level ");
                            data2(b[x+10],w4);
                        }
                    }
                    break;
                case 0x7D:  // Influence
                    w1 = *(unsigned short *) (&b[x+1]);
                    w2 = *(unsigned short *) (&b[x+6]);
                    fprintf(outFile,"Sim in ");
                    data2(b[x],w1);
                    fprintf(outFile,", output result to ");
                    if (b[x+5])
                        fprintf(outFile," stack object's object array: ");
                    else
                        fprintf(outFile," my object array: ");
                    if (readString2(gGroup, 0x118, w2) == 0)
                                fprintf(outFile,"[STR# 0x118:0x%X]", w2);
                    break;
            }
        }
    } else {

        // get function name and number of params if available

        i = readFn2(opcode);

        // interpret parameters for normal functions

        w2 = 0;
        if (i > 0) {
            if (*(unsigned long *) (&b[x]) != 0xFFFFFFFF) { // -1 = param killer
                fprintf(outFile,"(");
                if ((b[x+12] & 1) && (b[x+12] != 0xFF)) {     // param type
                    for (c1 = 1; c1 <= i; c1++) {
                        if (c1 != 1) fprintf(outFile,", ");
                        w3 = *(unsigned short *) (&b[x+w2+1]);
                        data2(b[x+w2], w3);                 // (owner, value), ...
                        w2 += 3;
                    }
                } else if ((b[x+12] & 2) && (b[x+12] != 0xFF) && nodeVersion) {
                    for (c1 = 0; c1 < i; c1++) {
                        if (c1 != 0) fprintf(outFile,", ");
                        fprintf(outFile,"original param %d",c1);
                    }
                } else {
                    for (c1 = 1; c1 <= i; c1++) {
                        if (c1 != 1) fprintf(outFile,", ");
                        w3 = *(unsigned short *) (&b[x+w2]);
                        if (w3 < 10)
                            fprintf(outFile,"%d", w3);      // literal values
                        else
                            fprintf(outFile,"%d (0x%X)", (short) w3, w3);
                        w2 += 2;
                    }
                }
                fprintf(outFile,")");
            }
        }
    }
}

//==============================================================================
// read out the code in the buffer

void read2(unsigned char *b, unsigned long s)
{
    unsigned char s2[STR_LENGTH], c0, c1, c2, c3;
    unsigned short f, op, w1, w2;
    unsigned long num, x, n2, d1;

    // get title (CString at 0)

    strncpy((char *) s2, (char *) b, 0x40);

    fprintf(outFile,"# Title = %s\n", s2);
    if (gNrefCrc != 0)
        fprintf(outFile,"# Filename: %s (%X)\n", gNref, gNrefCrc);
    if (gGlobGroup != gGroup)
        fprintf(outFile,"# SemiGlobal File: %s (%X)\n", gGlob, gGlobGroup);
    fprintf(outFile,"#\n");

    // get header

    x = 0;
    f = *(unsigned short *) (&b[0x40]);     // format
    num = *(unsigned short *) (&b[0x42]);   // # of opcodes
    c0 = b[0x44];                           // tree type
    c1 = b[0x45];                           // # of args
    c2 = b[0x46];                           // # of locals
    c3 = b[0x47];                           // header flag
    d1 = *(long *) (&b[0x48]);              // Tree version
    fprintf(outFile,"# Format = %X, Params = %d, Locals = %d\n", f, c1, c2);
    fprintf(outFile,"# Tree type = %X, Header flag = %X, Tree version = %X (%d)\n\n", c0, c3, d1, d1);
    x = 0x4C;

    // read instructions

    if (x > 0) {
        for (n2 = 0; n2 < num; n2++) {
            if (x >= s) {
                fprintf(outFile,"*** Error: too many instructions (%X) ***\n", num);
                break;
            }
            fprintf(outFile,"%6X: ", n2);                // line #
            op = *(unsigned short *) (&b[x]);   // opcode

            if (f < 0x8007) {
                c1 = b[x+2];            // true
                c2 = b[x+3];            // false
                if (f < 0x8005) {
                    readO2(op, b, x+4, f, 0);
                } else {
                    c3 = b[x+4];        // reserved
                    readO2(op, b, x+5, f, c3);
                }
                fprintf(outFile,"; true: ");
                readGx(c1);
                fprintf(outFile,", false: ");
                readGx(c2);
#if DEBUG_6 != 0
                if (op == DEBUG_6) {
                    if (f < 0x8005) {
                        fprintf(dFile6,"# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X\n",
                            b[x+4],b[x+5],b[x+6],b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],
                            b[x+12],b[x+13],b[x+14],b[x+15],b[x+16],b[x+17],b[x+18],b[x+19]);
                        fprintf(outFile,"\t# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X",
                            b[x+4],b[x+5],b[x+6],b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],
                            b[x+12],b[x+13],b[x+14],b[x+15],b[x+16],b[x+17],b[x+18],b[x+19]);
                    } else {
                        fprintf(dFile6,"# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X\n",
                            b[x+5],b[x+6],b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],b[x+12],
                            b[x+13],b[x+14],b[x+15],b[x+16],b[x+17],b[x+18],b[x+19],b[x+20]);
                        fprintf(outFile,"\t# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X",
                            b[x+5],b[x+6],b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],b[x+12],
                            b[x+13],b[x+14],b[x+15],b[x+16],b[x+17],b[x+18],b[x+19],b[x+20]);
                    }
                }
#endif
                fprintf(outFile,"\n");
                if (f < 0x8003)
                    x += 12;
                else if (f < 0x8005)
                    x += 20;
                else
                    x += 21;    // 5 ins bytes + 16 operand bytes
            }
            if (f == 0x8007) {
                w1 = *(unsigned short *) (&b[x+2]);     // true
                w2 = *(unsigned short *) (&b[x+4]);     // false
                c1 = b[x+6];                            // reserved
                readO2(op, b, x+7, f, c1);
                fprintf(outFile,"; true: ");
                readGx2(w1);
                fprintf(outFile,", false: ");
                readGx2(w2);
#if DEBUG_6 != 0
                if (op == DEBUG_6) {
                    fprintf(dFile6,"# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X\n",
                        b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],b[x+12],b[x+13],b[x+14],
                        b[x+15],b[x+16],b[x+17],b[x+18],b[x+19],b[x+20],b[x+21],b[x+22]);
                    fprintf(outFile,"\t# %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X %02X%02X",
                        b[x+7],b[x+8],b[x+9],b[x+10],b[x+11],b[x+12],b[x+13],b[x+14],
                        b[x+15],b[x+16],b[x+17],b[x+18],b[x+19],b[x+20],b[x+21],b[x+22]);
                }
#endif
                fprintf(outFile,"\n");
                x += 23;    // 7 ins bytes + 16 operand bytes
            }
        }
    }
}

//==============================================================================
// unpack the specified buffer

unsigned char *unPack(unsigned char *b, unsigned long s1, unsigned long s2)
{
    unsigned long x1, x2, i, o, q;
    unsigned char a, r, d;
    unsigned char *src, *dst;

    src = b;
    dst = (unsigned char *) malloc(s2);       // alloc destination buffer
    if (dst != NULL) {
        x1 = 9;     // src index
        x2 = 0;     // dest index

        // okay, now has some bounds checking

        while (x1 < s1) {

            // decode ctrl bytes

            a = src[x1++];                  // ctrl

            if ((a & 0x80) == 0) {          // 0ooqqqrr oooooooo
                r = a & 3;
                q = ((a >> 2) & 7) + 3;
                o = ((a << 3) & 0x300) + src[x1++];

            } else if ((a & 0x40) == 0) {   // 10qqqqqq rroooooo oooooooo
                q = (a & 0x3F) + 4;
                a = src[x1++];
                r = (a >> 6) & 3;
                o = ((a << 8) & 0x3F00) + src[x1++];

            } else if ((a & 0x20) == 0) {   // 110oqqrr oooooooo oooooooo qqqqqqqq
                r = a & 3;
                o = ((a << 12) & 0x10000) + (src[x1++] << 8);
                o += src[x1++];
                q = ((a << 6) & 0x300) + src[x1++] + 5;

            } else {                        // 111rrrrr / 111111rr (i'm pretty sure this will do it)
                r = ((a & 0x1F) + 1) << 2;
                if ((a & 0x1C) == 0x1C) r = a & 3;
                q = 0;
            }
#if 0
            if (gGroup == 0x7f277790 && src[7] == 0x24) printf("\n %X %X %X %X %X", x1, x2, r, q, o);
#endif
            // read block

            if (x2 + r <= s2) {
                for (i = 1; i <= r; i++) {      // read through a literal block
                    dst[x2++] = src[x1++];
                }
            } else {
                x1 += r;    // this will cause an assert failure (better than trashing memory :P)
                x2 += r;
            }

            // repeat block

            o = (x2 - o) - 1;                   // repeat index = current index (of last read byte) - offset
            if (x2 + q <= s2) {
                for (i = 1; i <= q; i++) {
                    dst[x2++] = dst[o++];
                }
            } else
                x2 += q;    // this will cause an assert failure
        }

        // index MUST = size at the end -- BIG trouble if these don't match up
#if 0
        if (x1 != s1 || x2 != s2) {
            printf("\nGroup = %X, %X %X %X %X\n", gGroup, x1, s1, x2, s2);
            FILE *df = fopen("dump2.dat", "wb");
            if (df != NULL) {
                fwrite(dst, s2, 1, df);
                fclose(df);
            }
            df = fopen("dump1.dat", "wb");
            if (df != NULL) {
                fwrite(src, s1, 1, df);
                fclose(df);
            }
        }
#endif
        assert(x1 == s1);
        assert(x2 == s2);
    }
    return dst;
}

//==============================================================================
// load and read data

unsigned char *get2(unsigned long o, unsigned long *s)
{
    FILE *f;
    unsigned long s1, s2;
    unsigned char *buf, *b2, cmp = 0;

    buf = NULL;
    s1 = *s;
    if ((f = fopen(gPack.name, "rb")) != NULL) {
        buf = (unsigned char *) malloc(s1);
        if (buf != NULL) {
            fseek(f, o, 0);
            fread(buf, s1, 1, f);       // get file data
            s2 = s1;
            b2 = buf; // b2 = buf

            // get compression header info (if any)

            if (s1 >= 10) {              // header + 1 ctrl
                s2 = *(unsigned long*) (&buf[0]);
                if (s2 == s1) {         // confirm size
                    if (b2[4] == 0x10 && b2[5] == 0xFB) {           // check cookie
                        s2 = (b2[6] << 16) + (b2[7] << 8) + b2[8];  // real size
                        cmp = 1;
                    }
                }
            }

            // get the data

            if (cmp) {
                b2 = unPack(buf, s1, s2);
                if (b2 != NULL) {
                    free(buf);  // discard old buf
                    buf = b2;   // buf = new buf
                    *s = s2;    // size = new size
                }
            }
        }
        fclose(f);
    }
    return buf;
}

//==============================================================================
// find specified BCON value and print it if available

void readConst2(unsigned short instance, unsigned short id)
{
    unsigned long s, v, x, *t, g;
    unsigned long eType, eGroup, eInstance, eOffset, eSize;
    unsigned char *buf;
    unsigned short w1, w2;

    // scan through the toc for this entry

    s = (gPack.size >> 2);
    v = gPack.ver;
    t = gPack.toc;
    g = gGroup;
    if (instance < 0x1000)
        g = 0x7FD46CD0;  // global
    else if (instance >= 0x2000 && instance < 0x3000 && gInstance >= 0x1000 && gInstance < 0x2000)
        g = gGlobGroup;

    x = 0;

    while (x < s) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;                     // skip subtype
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x42434F4E && eGroup == g && eInstance == instance) {
            // found
            buf = get2(eOffset, &eSize);    // get file into buffer
            if (buf != NULL) {
                w1 = 0x42 + (id + id);      // offset to value
                if ((w1+1) < eSize) {
                    w2 = *(unsigned short *) (&buf[w1]);
                    if (w2 < 16)
                        fprintf(outFile,"(%d)", w2);
                    else
                        fprintf(outFile,"(%d (0x%X))", (short) w2, w2);
                } else fprintf(outFile,"***");
                free(buf);                  // discard buffer
            }
        }
    }
}

//==============================================================================
// find specified STR# value and print it if available

int readString2(unsigned long group, unsigned long instance, unsigned short id)
{
    unsigned long s, v, x, *t, g, i;
    unsigned long eType, eGroup, eInstance, eOffset, eSize;
    unsigned char s2[STR_LENGTH];
    unsigned char *buf, ccode, c2;
    unsigned short w1, w2, w3;
    int found;

    // scan through the toc for this entry

    s = (gPack.size >> 2);
    v = gPack.ver;
    t = gPack.toc;
    i = instance;
    g = group;

    x = 0;
    found = 0;
    ccode = 1;          // just use US English for now

    while (x < s && found == 0) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;                     // skip subtype
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x53545223 && eGroup == g && eInstance == i) {
            // found
            buf = get2(eOffset, &eSize);    // get file into buffer
            if (buf != NULL) {
                w1 = *(unsigned short *) (&buf[0x42]);  // total num of strings
                w2 = 0x44;
                w3 = 0;
                while (w1 > 0) {
                    c2 = buf[w2++];         // get locale
                    if (c2 == ccode) {
                        if (w3 != id)
                            w3++;
                        else {
                            strcpy((char *) s2, (char *) &buf[w2]);
                            fprintf(outFile, "%s", s2); // value
                            found = 1;
                            w1 = 1;
                        }
                    }
                    while (buf[w2++] != 0); // skip value
                    while (buf[w2++] != 0); // skip description
                    w1--;
                }
                free(buf);                  // discard buffer
            }
        }
    }
    return found;
}


//==============================================================================
// find specified function and print its title if available

int readFn2(unsigned short o)
{
    unsigned long s, v, x, *t, g;
    unsigned long eType, eGroup, eInstance, eOffset, eSize;
    unsigned char s2[STR_LENGTH];
    unsigned char *buf;
    int p = -1;

    if (o < 0x100) {
        fprintf(outFile,"unknown primitive (0x%X)", o);
        p = 0;
    } else if (o <= LAST_GLOBAL) {
        x = o - 0x100;
        fprintf(outFile, "%s", gGlobalFNames[x]);  // global function name
        p = gGlobalArgs[x];                            // parameters
    }

    if (p < 0) {
        g = gGroup;                             // default to finding current group
        if (o >= 0x2000 && o < 0x3000 && gInstance >= 0x1000 && gInstance < 0x2000)
            g = gGlobGroup;

        // scan through the toc for this entry

        s = (gPack.size >> 2);
        v = gPack.ver;
        t = gPack.toc;
        x = 0;

        while (x < s && p < 0) {
            eType = t[x++];
            eGroup = t[x++];
            eInstance = t[x++];
            if (v > 1) x++;                     // skip subtype
            eOffset = t[x++];
            eSize = t[x++];

            if (eType == 0x42484156 && eGroup == g && eInstance == o) {

                // found

                buf = get2(eOffset, &eSize);    // get file into buffer
                if (buf != NULL) {
                    strncpy((char *) s2, (char *) buf, 0x40);
                    if (g != gGroup) fprintf(outFile, "%s", gGlob);     // calling semi
                    fprintf(outFile, ".%s", s2); // title
                    p = buf[0x45];              // # of args
                    free(buf);                  // discard buffer
                }
            }
        }

        // not found

        if (p < 0) {
            if (o < 0x1000) {
                fprintf(outFile,"global.0x%X", o);
            } else if (o < 0x2000) {
                fprintf(outFile,".0x%X", o);
            } else if (o < 0x3000) {
                if (strlen(gGlob) > 0)
                    fprintf(outFile,"%s.0x%X", gGlob, o);
                else
                    fprintf(outFile,"semiglobal.0x%X", o);
            } else {
                fprintf(outFile,"unknown.0x%X", o);
            }
        }
    }

    return p;
}

//==============================================================================
// find first BHAV group and count num of BHAVs

void findFirst()
{
    unsigned long s, v, x, *t;
    unsigned long eType, eGroup, eInstance, eOffset, eSize;
    unsigned char *buf, s0[STR_LENGTH];
    unsigned char c2;
    int found;


    // current = the lowest existing group or -1

    s = (gPack.size >> 2);
    v = gPack.ver;
    t = gPack.toc;
    gGroup = 0xFFFFFFFF;
    x = 0;
    gNum = 0;

    while (x < s) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x42484156) {
            gNum++;
            if (eGroup < gGroup) gGroup = eGroup;
        }
    }
    gInstance = 0;

    x = 0;
    found = 0;
    gNref[0] = 0;
    gNrefCrc = 0;

    while (x < s && found == 0) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;                     // skip subtype
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x4E524546 && eGroup == gGroup ) {
            // found
            buf = get2(eOffset, &eSize);    // get file into buffer
            if (buf != NULL) {
                strcpy((char *) gNref, (char *) &buf[0]);
                gNref[eSize] = 0;
                strcpy((char *) s0, (char *) gNref);
                gNrefCrc = crc_octets((char *) strlwr(s0), eSize) | 0x7F000000;
                free(buf);                  // discard buffer
                found = 1;
            }
        }
    }

    x = 0;
    found = 0;
    gGlob[0] = 0;
    gGlobGroup = gGroup;

    while (x < s && found == 0) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;                     // skip subtype
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x474C4F42 && eGroup == gGroup ) {
            // found
            buf = get2(eOffset, &eSize);    // get file into buffer
            if (buf != NULL) {
                c2 = buf[0x40];
                strcpy((char *) gGlob, (char *) &buf[0x41]);
                gGlob[c2] = 0;
                strcpy((char *) s0, (char *) gGlob);
                gGlobGroup = crc_octets((char *) strlwr(s0), c2) | 0x7F000000;
                free(buf);                  // discard buffer
                found = 1;
            }
        }
    }
}

//==============================================================================
// find next BHAV instance of the current group or the first of the next group

void findNext2(unsigned long g, unsigned long i, unsigned long *o1, unsigned long *s1)
{
    unsigned long s, v, x, *t, g2, g3, i2, i3, o2, s2;
    unsigned long eType, eGroup, eInstance, eOffset, eSize;
    unsigned char *buf, s0[STR_LENGTH];
    unsigned char c2;
    int found;


    s = (gPack.size >> 2);
    v = gPack.ver;
    t = gPack.toc;
    i2 = g3 = 0xFFFFFFFF;
    i3 = gInstance;
    g2 = gGroup;
    o2 = *o1;
    s2 = *s1;

    // find next instance in current group and next existing group

    x = 0;
    while (x < s) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) x++;
        eOffset = t[x++];
        eSize = t[x++];

        if (eType == 0x42484156) {
            if (eGroup == g2 && eInstance < i2 && eInstance > i3) {
                i2 = eInstance;
                o2 = eOffset;
                s2 = eSize;
            }
            if (eGroup < g3 && eGroup > g2)
                g3 = eGroup;
        }
    }

    // if no next instance exists, continue to next group...

    if (i2 == 0xFFFFFFFF) {

        gGroup = g3;

        // get first instance in this group

        x = 0;
        i2 = 0xFFFFFFFF;
        while (x < s) {
            eType = t[x++];
            eGroup = t[x++];
            eInstance = t[x++];
            if (v > 1) x++;
            eOffset = t[x++];
            eSize = t[x++];

            if (eType == 0x42484156 && eGroup == g3 && eInstance < i2) {
                i2 = eInstance;
                o2 = eOffset;
                s2 = eSize;
            }
        }

        x = 0;
        found = 0;
        gNref[0] = 0;
        gNrefCrc = 0;

        while (x < s && found == 0) {
            eType = t[x++];
            eGroup = t[x++];
            eInstance = t[x++];
            if (v > 1) x++;                     // skip subtype
            eOffset = t[x++];
            eSize = t[x++];

            if (eType == 0x4E524546 && eGroup == gGroup ) {
                // found
                buf = get2(eOffset, &eSize);    // get file into buffer
                if (buf != NULL) {
                    strcpy((char *) gNref, (char *) &buf[0]);
                    gNref[eSize] = 0;
                    strcpy((char *) s0, (char *) gNref);
                    gNrefCrc = crc_octets((char *) strlwr(s0), eSize) | 0x7F000000;
                    free(buf);                  // discard buffer
                    found = 1;
                }
            }
        }


        x = 0;
        found = 0;
        gGlob[0] = 0;
        gGlobGroup = gGroup;

        while (x < s && found == 0) {
            eType = t[x++];
            eGroup = t[x++];
            eInstance = t[x++];
            if (v > 1) x++;                     // skip subtype
            eOffset = t[x++];
            eSize = t[x++];

            if (eType == 0x474C4F42 && eGroup == gGroup ) {
                // found
                buf = get2(eOffset, &eSize);    // get file into buffer
                if (buf != NULL) {
                    if (buf[0]) {
                        c2 = buf[0x40];
                        strcpy((char *) gGlob, (char *) &buf[0x41]);
                        gGlob[c2] = 0;
                        strcpy((char *) s0, (char *) gGlob);
                        gGlobGroup = crc_octets((char *) strlwr(s0), c2) | 0x7F000000;
                    } else {
                        strcpy((char *) gGlob, (char *) &buf[0x40]);
                        strcpy((char *) s0, (char *) gGlob);
                        gGlobGroup = crc_octets((char *) strlwr(s0), strlen(s0)) | 0x7F000000;
                    }



                    free(buf);                  // discard buffer
                    found = 1;
                }
            }
        }
    }

    gInstance = i2;
    *o1 = o2;
    *s1 = s2;
}

//==============================================================================
// disassemble

void dis()
{
    unsigned long g, i, o, s, num;
    unsigned char *buf;
    FILE *f2;

    // setup progress output

    f2 = stderr;
    if (outFile != stdout) f2 = stdout;
    num = 0;

    // find first group and first instance in that group

    findFirst();
    findNext2(g, i, &o, &s);

//    while (gInstance != 0xFFFFFFFF) {
    while (num != gNum) {
        fprintf(outFile,"\n###\n# Group = %X, Instance = %X\n", gGroup, gInstance);

        buf = get2(o, &s);      // get file into buffer
        if (buf != NULL) {
            read2(buf, s);      // read buffer
            free(buf);          // discard buffer
        }
        num++;
#if 1
        if ((num & 0x0F) == 0 || num == gNum)  // progress
            fprintf(f2, "%d of %d BHAVs processed...\r", num, gNum);
#endif
        findNext2(g, i, &o, &s);    // next...
    }
}

//==============================================================================
// main

int main(int argc, char *argv[])
{
    gPack.toc = NULL;
    outFile = stdout;

#if DEBUG_6 != 0
    dFile6 = fopen("debug6.txt", "w");
#endif

    if (argc > 1) {
        if (argc > 2)
            outFile = fopen(argv[2], "w");      // 2nd param = output
        else
            outFile = stdout;                   // no param = print to console
        if (outFile == NULL) outFile = stdout;

        loadToc(argv[1]);
        if (gPack.toc != NULL) {
            dis();
            free(gPack.toc);     // discard toc
        }
    } else {
        printf("disaSim2 usage:\n\n\tdisaSim2 [package file][output file]\n");
        system("PAUSE");
    }
    if (outFile != stdout) fclose(outFile);

#if DEBUG_6 != 0
    if (dFile6 != NULL) fclose(dFile6);
#endif
    return 0;
}
